# Rekall Memory Forensics
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
The Rekall Entity Layer.
"""
__author__ = "Adam Sindelar <adamsh@google.com>"

import re

from rekall import registry

from rekall.entities.query import analyzer
from efilter import query as entity_query


class CostEnum(object):
    NoCost = 0
    LowCost = 1
    NormalCost = 2
    HighCost = 3
    VeryHighCost = 4


class EffectEnum(object):
    """Various outcomes of registering an entity."""

    Enqueued = 0  # Entity was enqueued for processing.
    Added = 1  # Result is a new entity.
    Merged = 2  # Result is an udpated (merged) entity.
    Duplicate = 3  # No new data.


class EntityCollector(object):
    """Base class for entity collectors.

    EntityCollector subclasses need to override the 'collect' method and the
    'outputs' ivar. EntityCollector takes care of registering collected
    entities with the EntityManager, deciding whether the collector should be
    called for manager-driven queries and ensuring collection is only done once
    per cache lifetime.

    Intended subclass overrides:
    ============================

    ### self.outputs (ivar, required):
    Override with a list of collection promises. Each promise should represent
    a component the collector yields on a call to collect, like so:

    outputs = ["Process", "Timestamps"]

    The collector may also specify a value it guarantees will be set on the
    entities it yields. The format for doing so is:

    outputs = ["Struct/type=proc"]

    The collector is not required to actually deliver on all of its promises,
    but it is not allowed to yield something it didn't promise.

    ### self.collect_args (ivar, optional):
    Collectors can specify the kind of input they want using a dictionary of
    queries describing the input, keyed on the names of kwargs they should be
    supplied as. The manager may call collectors with dependencies more than
    once, as new data becomes available.

    Examples:

    # ProcessParser wants proc structs. Its collect method will now receive
    # a keyword arg 'procs' populated with entities matching the query.
    collect_args = {"procs": "Struct/type is 'proc'"}

    # Socket/process relationship inference wants sockets and processes.
    # Its collect method will now receive two keyword args - 'processes' with
    # process entities, and 'sockets' with base object entities.
    collect_args = {"processes": "has component Process",
                    "sockets": "Struct/type" is 'socket'}

    ### self.filter_input (ivar, optional):
    Costly collectors can flip this variable to True, which will cause the
    manager to call self.input_filter to prefilter the ingestion set, giving
    the collector the opportunity to filter out entities it has parsed before.

    ### self.run_cost (ivar, optional):
    Should be set to a CostEnum value, estimating how expensive the collector is
    in terms of performance.

    ### self.enforce_hint (ivar, optional):
    If True, the manager will always supply a hint to the collect function. If
    False, the manager will only supply a hint when it's collecting for
    artifacts.

    ### self.complete_input (ivar, optional):
    If True, will cause the manager to always call the collect method with all
    available results for the ingestion queries, even if they've been sent
    before and haven't been updated since. The results can still be filtered
    with filter_input.
    """

    outputs = []  # Subclasses must override. See above.
    _promises = None  # Promises (SimpleDependency) generated from outputs.

    collect_args = None  # Subclasses may override.
    collect_queries = None  # Will be populated automatically on init.

    run_cost = CostEnum.NormalCost
    enforce_hint = False
    filter_input = False
    complete_input = False

    __metaclass__ = registry.MetaclassRegistry
    __abstract = True

    DELIMITER = re.compile(r"[\/=]")

    def __init__(self, entity_manager=None):
        self.manager = entity_manager
        self.session = entity_manager.session
        self._indices_seen = set()
        self.collect_queries = {}
        self._ensure_compile_queries()

    @property
    def profile(self):
        return self.session.profile

    @property
    def name(self):
        return getattr(self, "_name", type(self).__name__)

    @property
    def promises(self):
        self._ensure_compile_promises()
        return self._promises

    def _ensure_compile_promises(self):
        if not self._promises:
            self._promises = [analyzer.SimpleDependency.parse(output)
                              for output in self.outputs]

        # All collectors return Entity.
        self._promises.append(analyzer.SimpleDependency("Entity"))

    def _ensure_compile_queries(self):
        if self.collect_queries or self.collect_args is None:
            return

        for arg, source in self.collect_args.iteritems():
            self.collect_queries[arg] = entity_query.Query(source)

    @property
    def is_collected(self):
        """Set to True after self.collect runs with no hints.

        Running collect with a non-null hint will not flip this."""
        return self.name in self.manager.finished_collectors

    # pylint: disable=unused-argument
    def input_filter(self, hint, entities=None):
        """Filter the ingest set. Use to prevent parsing the same thing twice.

        Default implementation of the ingest filter will keep a set of entities
        it has processed before and filter those out. Subclasses can override.

        NOTE: Ingest filter is disabled by default - subclasses that wish the
        manager to enable it must signal so by setting filter_input to True.
        """
        for entity in entities:
            if not entity.indices & self._indices_seen:
                yield entity

            self._indices_seen |= entity.indices

    def prebuild(self, components, keys):
        """Prebuilds identity based on data in components.

        In the olden days, identities could be arbitrary and were not tied
        to data in the entity. The current API (Manager.identify) theoretically
        still allows creation of identities with arbitrary indices, but that's
        slow, as the data in the entity and the identity then has to be
        type-coerced separately. This method is the new API and only coerces
        data once.

        Arguments:
            components: Instances of Component that the collector is going to
                        yield.
            keys: List of attributes that will be used to build the identity.

        Returns tuple of:
        - The identity.
        - List that can be yielded from Collector.collect, which includes the
          identity.
        """
        index_values = []
        for key in keys:
            component_name, field_name = key.split("/", 1)

            # Find the component.
            for component in components:
                if component.component_name == component_name:
                    index_values.append(component[field_name])
                    break

        identity = self.manager.identify_no_cast(
            {tuple(keys): tuple(index_values)})
        return identity, [identity] + list(components)

    def collect(self, hint):
        """Override to yield components - analogous to 'calculate', but typed.

        Subclasses should override this to yield components that represent the
        data they collect. For example, a pslist collector would yield Process
        components. Only components listed in self.outputs should be yielded
        and no other.

        Output format:
        ==============

        In the simplest case, yield one component at a time. To indicate two
        (or more) components are related, yield both in a list (for example,
        Process and the Struct wrapping the proc struct representing it
        in kernel).

        Unless an identity is specified explicitly, one will be created
        automatically by taking the first field of the first component and
        instantiating an identity with it.

        To specify an identity explicitly, yield the Entity component as first
        in the list and provide the identity in its 'identity' field.

        Examples:
        =========

        # Yielding just one component at a time:
        yield manager.Process(
            pid=proc.p_pid,
            ...)

        # Yielding the process component and the BSD proc struct:
        yield [
            manager.Process(pid=proc.p_pid, ...),
            manager.Struct(base=proc, ...),
        ]

        # In both above examples, the identity will be "Process/pid=(PID here)."
        # Specifying an explicit identity would look like this:
        yield [
            manager.Entity(identity=ProcessIdentity(...)),
            manager.Process(pid=proc.p_pid, ...),
            manager.Struct(base=proc, ...),
        ]

        Optimization hints:
        ===================

        A hint may be passed to the collect method, in which case the collector
        may elect to only collect some information it's capable of collecting
        and not all. For example, a hint to the pslist collector may specify
        that the caller only cares about processes listed by walking the
        process head list. The format of the hint is currently not specified.
        """
        pass

    @classmethod
    def is_active(cls, session):
        return cls.run_cost <= session.GetParameter("max_collector_cost")

    @property
    def entities(self):
        """Returns this collector's entities."""
        return self.session.entities.find_by_collector(self.name)



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
@author:       Andrew Case
@license:      GNU General Public License 2.0 or later
@contact:      atcuno@gmail.com
@organization: Digital Forensics Solutions
"""
import itertools

from rekall.plugins.linux import common


class Ifconfig(common.LinuxPlugin):
    '''Gathers active interfaces.'''

    __name = "ifconfig"

    def enumerate_devices(self):
        """A generator over devices.

        Yields:
          a tuple of (name, ip_addr, mac_addr, promisc).
        """
        return itertools.chain(self.get_devs_namespace(),
                               self.get_devs_base())

    def get_devs_base(self):
        net_device = self.profile.get_constant_object(
            "dev_base", target="net_device", vm=self.kernel_address_space)

        for net_dev in net_device.walk_list("next"):
            yield net_dev

    def gather_net_dev_info(self, net_dev):
        mac_addr = net_dev.mac_addr

        for dev in net_dev.ip_ptr.ifa_list.walk_list("ifa_next"):
            yield dev.ifa_label, dev.ifa_address, mac_addr, net_dev.flags

    def get_devs_namespace(self):
        nethead = self.profile.get_constant_object(
            "net_namespace_list", target="list_head",
            vm=self.kernel_address_space)

        for net in nethead.list_of_type("net", "list"):
            for net_dev in net.dev_base_head.list_of_type("net_device",
                                                          "dev_list"):
                yield net_dev

    def render(self, renderer):
        renderer.table_header([("Interface", "interface", "<16"),
                               ("Ipv4Address", "ipv4", "<20"),
                               ("MAC", "mac", "18"),
                               ("Flags", "flags", "<20")])

        for net_dev in self.enumerate_devices():
            for name, ipv4, mac, flags in self.gather_net_dev_info(net_dev):
                renderer.table_row(name, ipv4, mac, flags)



# Rekall Memory Forensics
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
"""Enumerate and dump all compressed memory pages on Darwin."""

__author__ = "Andreas Moser <amoser@google.com>"

import os

from rekall.plugins import core
from rekall.plugins.darwin import common
from rekall.plugins.darwin import WKdm


class DarwinDumpCompressedPages(core.DirectoryDumperMixin, common.DarwinPlugin):
    """Dumps all compressed pages."""

    __name = "dumpcompressedmemory"

    SLOT_ARRAY_SIZE = 64
    PAGE_SIZE = 4096

    def UnpackCSize(self, c_slot):

        size = c_slot.c_size
        if size == self.PAGE_SIZE - 1:
            return self.PAGE_SIZE
        else:
            return size

    def render(self, renderer):

        pages = self.profile.get_constant_object("_c_segment_count", "int")

        renderer.format("Going to dump {0} segments.\n", pages)

        p_segu = self.profile.get_constant_object(
            "_c_segments", "Pointer", target_args={
                "target": "Array",
                "target_args": {
                    "target": "c_segu",
                "count": int(pages),
                }})

        segu_array = p_segu.deref()

        for i, segu in enumerate(segu_array):
            renderer.RenderProgress("Segment: %d" % i)

            c_seg = segu.c_seg

            if (c_seg.c_ondisk or
                c_seg.c_on_swappedout_q or
                c_seg.c_on_swappedout_sparse_q):
                # Data swapped out.
                continue

            c_buffer = c_seg.c_store.c_buffer
            if c_buffer == 0:
                # No data in this segment.
                continue

            seg_buffer = c_buffer.obj_vm.read(c_buffer.v(),
                                              c_seg.c_nextoffset * 4)

            c_slot_arrays = []
            for slot in c_seg.c_slots:
                c_slot_arrays.append(
                    slot.dereference_as(
                        target="Array", target_args=dict(target="c_slot")))

            for slot_nr in xrange(c_seg.c_nextslot):
                c_slot_array = c_slot_arrays[slot_nr / self.SLOT_ARRAY_SIZE]
                c_slot = c_slot_array[slot_nr % self.SLOT_ARRAY_SIZE]

                if not (c_slot.c_offset and c_slot.c_size):
                    continue

                c_size = self.UnpackCSize(c_slot)

                # This should never happen.
                if c_slot.c_offset * 4 + c_size >= len(seg_buffer):
                    continue

                data = seg_buffer[c_slot.c_offset * 4:
                                  c_slot.c_offset * 4 + c_size]

                offset_alignment_mask = 0x3

                c_rounded_size = (c_size + offset_alignment_mask)
                c_rounded_size &= ~offset_alignment_mask

                if (c_rounded_size == self.PAGE_SIZE):
                    # Page was not compressible.
                    # Copy anyways?

                    # with renderer.open(
                    #         directory=self.dump_dir,
                    #         filename="seg%d_slot%d.dat" % (i, slot_nr),
                    #         mode="wb") as fd:
                    #     fd.write(data)
                    continue

                try:
                    decompressed = WKdm.WKdm_decompress_apple(data)
                    if decompressed:
                        dirname = os.path.join(self.dump_dir, "segment%d" % i)
                        try:
                            os.mkdir(dirname)
                        except OSError:
                            pass

                        with renderer.open(
                                directory=dirname,
                                filename="slot%d.dmp" % slot_nr,
                                mode="wb") as fd:
                            fd.write(decompressed)

                except Exception as e:
                    renderer.report_error(str(e))




# Rekall Memory Forensics
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""The plugins in this module are mainly used to visually test renderers."""

__author__ = "Adam Sindelar <adamsh@google.com>"

import itertools

from rekall import plugin
from rekall import algo
from rekall import utils

from rekall.plugins.renderers import visual_aides


class RekallBovineExperience3000(plugin.Command):
    """Renders Bessy the cow and some beer.

    This is a text renderer stress-test. It uses multiple features at the
    same time:

    - Multiple coloring rules per line (this was a doozy).
    - Two columns with colors next to each other.
    - Text with its own newlines isn't rewrapped.
    - It still wraps if it overflows the cell.
    - Bovine readiness and international spirit.
    """
    __name = "moo"

    def render(self, renderer):
        renderer.table_header([
            dict(name="Dogma", width=35, style="full"),
            dict(name="Bessy", width=65, type="bool", style="cow"),
            dict(name="Pilsner", width=50, style="full"),
            dict(name="Nowrap", width=10, nowrap=True)])

        fixtures = self.session.LoadProfile("tests/fixtures")
        beer = fixtures.data["ascii_art"]["beer"]
        phys_map = fixtures.data["fixtures"]["phys_map"]

        renderer.table_row(
            ("This is a renderer stress-test. The flags should have correct"
             " colors, the beer should be yellow and the cell on the left"
             " should not bleed into the cell on the right.\n"
             "This is a really "
             "long column of text with its own newlines in it!\n"
             "This bovine experience has been brought to you by Rekall."),
            True,
            utils.AttributedString("\n".join(beer["ascii"]),
                                   beer["highlights"]),
            ("This is a fairly long line that shouldn't get wrapped.\n"
             "The same row has another line that shouldn't get wrapped."))

        renderer.section("Heatmap test:")
        cells = []
        for digit in itertools.islice(algo.EulersDecimals(), 0xff):
            cells.append(dict(heat=float(digit + 1) * .1, value=digit))

        randomized = visual_aides.Heatmap(
            caption="Offset (p)",
            # Some of the below xs stand for eXtreme. The other ones just
            # look cool.
            column_headers=["%0.2x" % x for x in xrange(0, 0xff, 0x10)],
            row_headers=["0x%0.6x" % x for x
                         in xrange(0x0, 0xfffff, 0x10000)],
            cells=cells,
            greyscale=False)

        gradual = visual_aides.Heatmap(
            caption="Offset (v)",
            column_headers=["%0.2x" % x for x in xrange(0, 0xff, 0x10)],
            row_headers=["0x%0.6x" % x for x
                         in xrange(0x0, 0xfffff, 0x10000)],
            cells=[dict(value="%x" % x, heat=x / 255.0) for x in xrange(256)],
            greyscale=False)

        ranges_legend = visual_aides.MapLegend(phys_map["ranges_legend"])

        ranges = visual_aides.RunBasedMap(
            caption="Offset (p)",
            legend=ranges_legend,
            runs=phys_map["runs"])

        renderer.table_header([dict(name="Random Heatmap", style="full",
                                    width=60, align="c"),
                               dict(name="Gradual Heatmap", style="full",
                                    width=60, align="c"),
                               dict(name="Legend", style="full",
                                    orientation="horizontal")])
        renderer.table_row(randomized, gradual, visual_aides.HeatmapLegend())

        renderer.table_header([dict(name="Greyscale Random", style="full",
                                    width=60, align="c"),
                               dict(name="Memory Ranges", style="full",
                                    width=80, align="c"),
                               dict(name="Ranges Legend", style="full",
                                    width=30, orientation="vertical")])

        randomized.greyscale = True
        renderer.table_row(randomized, ranges, ranges_legend)



# Rekall Memory Forensics
# Copyright (C) 2012 Michael Cohen
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""Plugins allow the core rekall system to be extended."""

__author__ = "Michael Cohen <scudette@gmail.com>"


import StringIO

from rekall import config
from rekall import obj
from rekall import registry
from rekall.ui import text as text_renderer


class Error(Exception):
    """Raised for plugin errors."""


class PluginError(Error):
    """An error occured in a plugin."""


class InvalidArgs(Error):
    """Invalid arguments."""


class Abort(Error):
    """Signal aborting of the plugin."""


class Command(object):
    """A command can be run from the rekall command line.

    Commands can be automatically imported into the shell's namespace and are
    expected to produce textual (or other) output.

    In order to define a new command simply extend this class.
    """

    # these attribute are not inherited.

    # The name of this command (The command will be registered under this
    # name). If empty, the command will not be imported into the namespace but
    # will still be available from the Factory below.
    __name = ""

    # Name of the category of this command. This is used when showing help and
    # in the UI.
    __category = ""

    # This class will not be registered (but extensions will).
    __abstract = True
    __metaclass__ = registry.MetaclassRegistry

    # This declares that this plugin only exists in the interactive session.
    interactive = False

    # This will hold the error status from running this plugin.
    error_status = None

    @classmethod
    def args(cls, parser):
        """Declare the command line args this plugin needs."""

    @registry.classproperty
    def name(cls):  # pylint: disable=no-self-argument
        return getattr(cls, "_%s__name" % cls.__name__, None)

    def __init__(self, **kwargs):
        """The constructor for this command.

        Commands can take arbitrary named args and have access to the running
        session.

        Args:
          session: The session we will use. Many options are taken from the
            session by default, if not provided. This allows users to omit
            specifying many options.
        """
        session = kwargs.pop("session", None)
        if kwargs:
            raise InvalidArgs("Invalid arguments: %s" % unicode(kwargs.keys()))

        super(Command, self).__init__(**kwargs)

        if session == None:
            raise InvalidArgs("A session must be provided.")

        self.session = session

    def get_plugin(self, name, **kwargs):
        """Returns an instance of the named plugin.

        The new plugin will initialized with the current session and optional
        kwargs.
        Args:
          name: The generic name of the plugin (i.e. the __name attribute,
             e.g. pslist).
          kwargs: Extra args to use for instantiating the plugin.
        """
        for cls in self.classes.values():
            if cls.name == name and cls.is_active(self.session):
                return cls(session=self.session, profile=self.profile,
                           **kwargs)

    def __str__(self):
        """Render into a string using the text renderer."""
        fd = StringIO.StringIO()
        ui_renderer = text_renderer.TextRenderer(
            session=self.session, fd=fd)

        with ui_renderer.start(plugin_name=self.name):
            self.render(ui_renderer)

        return fd.getvalue()

    def __repr__(self):
        return "Plugin: %s" % self.name

    def render(self, renderer):
        """Produce results on the renderer given.

        Each plugin should implement this method to produce output on the
        renderer. The framework will initialize the plugin and provide it with
        some kind of renderer to write output on. The plugin should not assume
        that the renderer is actually TextRenderer, only that the methods
        defined in the BaseRenderer exist.

        Args:
          renderer: A renderer based at rekall.ui.renderer.BaseRenderer.
        """

    @classmethod
    def is_active(cls, session):
        """Checks we are active.

        This method will be called with the session to check if this specific
        class is active. This mechanism allows multiple implementations to all
        share the same name, as long as only one is actually active. For
        example, we can have a linux, windows and mac version of plugins with
        the "pslist" name.
        """
        _ = session
        return True

    @classmethod
    def GetActiveClasses(cls, session):
        """Return only the active commands based on config."""
        for command_cls in cls.classes.values():
            if command_cls.is_active(session):
                yield command_cls


class ProfileCommand(Command):
    """A baseclass for all commands which require a profile."""

    __abstract = True

    PROFILE_REQUIRED = True

    @classmethod
    def args(cls, metadata):
        # Top level args.
        metadata.add_argument(
            "-p", "--profile", critical=True,
            help="Name of the profile to load. This is the "
            "filename of the profile found in the profiles "
            "directory. Profiles are searched in the profile "
            "path order.")

        metadata.add_requirement("profile")

    @classmethod
    def is_active(cls, session):
        if cls.PROFILE_REQUIRED:
            # Note! This will trigger profile autodetection if this plugin is
            # needed. This might be slightly unexpected: When command line
            # completing the available plugins we will trigger profile
            # autodetection in order to determine which plugins are active.
            profile = (super(ProfileCommand, cls).is_active(session) and
                       session.profile != None)

            return profile

        else:
            return True

    def __init__(self, profile=None, **kwargs):
        """Baseclass for all plugins which accept a profile.

        Args:
          profile: The kernel profile to use for this command.
        """
        super(ProfileCommand, self).__init__(**kwargs)

        # If a profile was provided we must set it into the session and then use
        # it. (The new profile must control the presence of other dependent
        # plugins and so forms part of the session's state.).
        if profile is not None:
            self.session.profile = profile

        # If the session already has a profile, use it.
        if self.session.HasParameter("profile_obj"):
            self.profile = self.session.profile

        # If the profile is required but the session has nothing yet, force
        # autodetection.
        elif self.PROFILE_REQUIRED:
            # Force autodetection...
            self.profile = self.session.profile

            # Nothing found... bail out!
            if not self.profile:
                raise PluginError(
                    "Profile could not detected. "
                    "Try specifying one explicitly.")


class KernelASMixin(object):
    """A mixin for those plugins which require a valid kernel address space.

    This class ensures a valid kernel AS exists or an exception is raised.
    """

    @classmethod
    def args(cls, parser):
        """Declare the command line args we need."""
        super(KernelASMixin, cls).args(parser)

        parser.add_argument("--dtb", type="IntParser",
                            help="The DTB physical address.")

    def __init__(self, dtb=None, **kwargs):
        """A mixin for plugins which require a valid kernel address space.

        Args:
          dtb: A potential dtb to be used.
        """
        super(KernelASMixin, self).__init__(**kwargs)

        # If the dtb is specified use that as the kernel address space.
        if dtb is not None:
            self.kernel_address_space = (
                self.session.kernel_address_space.__class__(
                    base=self.physical_address_space, dtb=dtb))
        else:
            # Try to load the AS from the session if possible.
            self.kernel_address_space = self.session.kernel_address_space

        if self.kernel_address_space == None:
            # Try to guess the AS
            self.session.plugins.load_as().GetVirtualAddressSpace()

            self.kernel_address_space = self.session.kernel_address_space

        if self.kernel_address_space == None:
            raise PluginError("kernel_address_space not specified.")


class PhysicalASMixin(object):
    """A mixin for those plugins which require a valid physical address space.

    This class ensures a valid physical AS exists or an exception is raised.
    """

    PHYSICAL_AS_REQUIRED = True

    @classmethod
    def args(cls, metadata):
        super(PhysicalASMixin, cls).args(metadata)
        metadata.add_requirement("physical_address_space")

    def __init__(self, **kwargs):
        """A mixin for those plugins requiring a physical address space.

        Args:
          physical_address_space: The physical address space to use. If not
            specified we use the following options:

            1) session.physical_address_space,

            2) Guess using the load_as() plugin,

            3) Use session.kernel_address_space.base.

        """
        super(PhysicalASMixin, self).__init__(**kwargs)
        self.physical_address_space = self.session.physical_address_space

        if not self.physical_address_space:
            # Try to guess the AS
            self.session.plugins.load_as().GetPhysicalAddressSpace()
            self.physical_address_space = self.session.physical_address_space

        if self.PHYSICAL_AS_REQUIRED and not self.physical_address_space:
            raise PluginError("Physical address space is not set. "
                              "(Try plugins.load_as)")


class VerbosityMixIn(object):
    """Use this mixin to provide a --verbosity option to a plugin."""

    @classmethod
    def args(cls, parser):
        super(VerbosityMixIn, cls).args(parser)

        parser.add_argument(
            "-V", "--verbosity", default=1, type="IntParser",
            help="An integer reflecting the amount of desired output: "
            "0 = quiet, 10 = noisy. Default: 1")

    def __init__(self, *args, **kwargs):
        # Do not interfere with positional args, since this is a mixin.
        self.verbosity = kwargs.pop("verbosity", 1)
        super(VerbosityMixIn, self).__init__(*args, **kwargs)



class PluginMetadataDatabase(object):
    """A database of all the currently registered plugin's metadata."""

    def __init__(self, session):
        self.session = session
        self.Rebuild()

    def Rebuild(self):
        self.db = {}

        for plugin_cls in Command.classes.itervalues():
            plugin_name = plugin_cls.name
            self.db.setdefault(plugin_name, []).append(
                config.CommandMetadata(plugin_cls))

    def MetadataByName(self, name):
        """Return all Implementations that implement command name."""
        for command_metadata in self.db[name]:
            yield command_metadata

    def GetActivePlugin(self, plugin_name):
        results = []
        for command_metadata in self.db.get(plugin_name, []):
            plugin_cls = command_metadata.plugin_cls
            if plugin_cls.is_active(self.session):
                results.append(command_metadata)

        # We assume there can only be one active plugin implementation. It
        # is an error to have multiple implementations active at the same
        # time.
        if len(results) > 1:
            raise RuntimeError("Multiple plugin implementations for %s: %s" % (
                plugin_name, results))

        if results:
            return results[0]

        return obj.NoneObject("Plugin not active")

    def Serialize(self):
        result = {}
        for name in self.db:
            command_metadata = self.GetActivePlugin(name)
            if command_metadata:
                result[name] = command_metadata.Metadata()

        return result

    def GetRequirments(self, command_name):
        result = set()
        for metadata in self.db[command_name]:
            result.update(metadata.requirements)

        return result



# Rekall Memory Forensics
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""This file provides read/write support for EWF files.

EWF files are generated by Encase/FTK and are a common compressible storage
format for digital evidence.

The below code is based on libewf:
https://github.com/libyal/libewf
https://googledrive.com/host/0B3fBvzttpiiSMTdoaVExWWNsRjg/


NOTE: Since EWFv1 files are unable to represent sparse data they are not
directly suitable for storing memory images. Therefore in Rekall we generally
use EWF files as containers for other formats, such as ELF core dumps.

NOTE: EWF files produced by the ewfacquire plugin are _NOT_ compatible with
Encase/FTK and can not be analyzed by those programs. We merely use the EWF
container as a container providing seekable compression for more traditional
memory image formats such as ELF.

When using the ewfacquire plugin, if the source address space contains a single
run of data, we generate a single EWF file of this run (e.g. for a disk
image). If, however, the source address space contains more than one run, we
automatically create an ELF core dump to contain the sparse runs, and that is
compressed into the EWF file instead. This is not generally compatible with
Encase or FTK since they do not understand layered address spaces! For Rekall
this works because Rekall automatically detects that the EWF file contains an
ELF core dump and stacks the relevant address spaces.
"""

__author__ = "Michael Cohen <scudette@google.com>"
import array
import os
import struct
import zlib

from rekall import obj
from rekall import plugin
from rekall import testlib
from rekall import utils
from rekall.plugins.addrspaces import elfcore
from rekall.plugins.addrspaces import standard
from rekall.plugins.overlays import basic


EWF_TYPES = dict(
    ewf_file_header_v1=[0x0d, {
        'EVF_sig': [0, ['Signature', dict(value="EVF\x09\x0d\x0a\xff\x00")]],

        'fields_start': [8, ['byte']],
        'segment_number': [9, ['unsigned short int']],
        'fields_end': [11, ['unsigned short int']],
        }],

    ewf_file_header_v2=[None, {
        'EVF_sig': [0, ['Signature', dict(value="EVF2\x0d\x0a\x81\x00")]],

        'major_version': [9, ['byte']],
        'minor_version': [10, ['byte']],

        'compression_method': [11, ['Enumeration', dict(
            target="unsigned short int",
            choices=dict(
                NONE=0,
                DEFLATE=1,
                BZIP2=2,
                )
            )]],
        'segment_number': [13, ['unsigned short int']],
        'set_identifier': [15, ['String', dict(length=16)]],
        }],

    ewf_section_descriptor_v1=[76, {
        # This string determines how to process this section.
        'type': [0, ['String', dict(length=16)]],

        # The next section in this file.
        'next': [16, ['Pointer', dict(
            target="ewf_section_descriptor_v1"
            )]],

        'size': [24, ['long long unsigned int']],
        'checksum': [72, ['int']],
        }],

    ewf_volume=[94, {
        'media_type': [0, ['Enumeration', dict(
            choices={
                0: 'remobable_disk',
                1: 'fixed_disk',
                2: 'optical_disk',
                3: 'LVF',
                4: 'memory',
                },
            )]],
        'number_of_chunks': [4, ['unsigned int']],
        'sectors_per_chunk': [8, ['unsigned int']],
        'bytes_per_sector': [12, ['unsigned int']],
        'number_of_sectors': [16, ['long long unsigned int']],
        'chs_cylinders': [24, ['unsigned int']],
        'chs_heads': [28, ['unsigned int']],
        'chs_sectors': [32, ['unsigned int']],

        'media_flags': [36, ['Flags', dict(
            maskmap={
                'image': 1,
                'physical': 2,
                'Fastblock Tableau write blocker': 4,
                'Tableau write blocker': 8
                })]],

        'compression_level': [52, ['Enumeration', dict(
            choices={
                0: 'no compression',
                1: 'fast/good compression',
                2: 'best compression',
                })]],

        'checksum': [90, ['int']],
        }],

    ewf_table_entry=[4, {
        # Is the chunk compressed?
        'compressed': [0, ['BitField', dict(start_bit=31, end_bit=32)]],

        # The offset to the chunk within the file.
        'offset': [0, ['BitField', dict(start_bit=0, end_bit=31)]],
        }],

    ewf_table_header_v1=[lambda x: x.entries[x.number_of_entries].obj_end, {
        'number_of_entries': [0, ['long long unsigned int']],
        'base_offset': [8, ['long long unsigned int']],
        'checksum': [20, ['int']],

        # The table just contains a list of table entries to the start of each
        # chunk.
        'entries': [24, ['Array', dict(
            target='ewf_table_entry',
            count=lambda x: x.number_of_entries
            )]],
        }],

    )


class ewf_section_descriptor_v1(obj.Struct):
    def UpdateChecksum(self):
        """Recalculate the checksum field."""
        self.size = self.next.v() - self.obj_offset
        data = self.obj_vm.read(
            self.obj_offset, self.checksum.obj_offset - self.obj_offset)

        self.checksum = zlib.adler32(data)


class ewf_table_header_v1(obj.Struct):
    def UpdateChecksum(self):
        """Recalculate the checksum field."""
        data = self.obj_vm.read(
            self.obj_offset, self.checksum.obj_offset - self.obj_offset)

        self.checksum = zlib.adler32(data)


class ewf_volume(ewf_table_header_v1):
    pass


class EWFProfile(basic.ProfileLLP64, basic.BasicClasses):
    """Basic profile for EWF files."""

    @classmethod
    def Initialize(cls, profile):
        super(EWFProfile, cls).Initialize(profile)

        profile.add_types(EWF_TYPES)
        profile.add_classes(
            ewf_section_descriptor_v1=ewf_section_descriptor_v1,
            ewf_table_header_v1=ewf_table_header_v1,
            ewf_volume=ewf_volume,
            )


class EWFFile(object):
    """A helper for parsing an EWF file."""

    def __init__(self, session=None, address_space=None):
        self.session = session

        # This is a cache of tables. We can quickly find the table responsible
        # for a particular chunk.
        self.tables = utils.SortedCollection(key=lambda x: x[0])
        self._chunk_offset = 0
        self.chunk_size = 32 * 1024

        # 32kb * 100 = 3.2mb cache size.
        self.chunk_cache = utils.FastStore(max_size=100)

        self.address_space = address_space
        self.profile = EWFProfile(session=session)
        self.file_header = self.profile.ewf_file_header_v1(
            offset=0, vm=self.address_space)

        # Make sure the file signature is correct.
        if not self.file_header.EVF_sig.is_valid():
            raise RuntimeError("EVF signature does not match.")

        # Now locate all the sections in the file.
        first_section = self.profile.ewf_section_descriptor_v1(
            vm=self.address_space, offset=self.file_header.obj_end)

        for section in first_section.walk_list("next"):
            if section.type == "header2":
                self.handle_header2(section)

            elif section.type == "header":
                self.handle_header(section)

            elif section.type in ["disk", "volume"]:
                self.handle_volume(section)

            elif section.type == "table":
                self.handle_table(section)

        # How many chunks we actually have in this file.
        self.size = self._chunk_offset * self.chunk_size

    def handle_header(self, section):
        """Handle the header section.

        We do not currently do anything with it.
        """
        # The old header contains an ascii encoded description, compressed with
        # zlib.
        data = zlib.decompress(
            section.obj_vm.read(section.obj_end, section.size))

        # We dont do anything with this data right now.

    def handle_header2(self, section):
        """Handle the header2 section.

        We do not currently do anything with it.
        """
        # The header contains a utf16 encoded description, compressed with zlib.
        data = zlib.decompress(
            section.obj_vm.read(section.obj_end, section.size)).decode("utf16")

        # We dont do anything with this data right now.

    def handle_volume(self, section):
        """Handle the volume section.

        We mainly use it to know the chunk size.
        """
        volume_header = self.profile.ewf_volume(
            vm=self.address_space, offset=section.obj_end)

        self.chunk_size = (volume_header.sectors_per_chunk *
                           volume_header.bytes_per_sector)

    def handle_table(self, section):
        """Parse the table and store it in our lookup table."""
        table_header = self.profile.ewf_table_header_v1(
            vm=self.address_space, offset=section.obj_end)

        number_of_entries = table_header.number_of_entries

        # This is an optimization which allows us to avoid small reads for each
        # chunk. We just load the entire table into memory and read it on demand
        # from there.
        table = array.array("I")
        table.fromstring(self.address_space.read(
            table_header.entries.obj_offset,
            4 * table_header.number_of_entries))

        # We assume the last chunk is a full chunk. Feeding zlib.decompress()
        # extra data does not matter so we just read the most we can.
        table.append(table[-1] + self.chunk_size)

        self.tables.insert(
            # First chunk for this table, table header, table entry cache.
            (self._chunk_offset, table_header, table))

        # The next table starts at this chunk.
        self._chunk_offset += number_of_entries

    def read_chunk(self, chunk_id):
        """Read a single chunk from the file."""
        try:
            return self.chunk_cache.Get(chunk_id)
        except KeyError:
            start_chunk, table_header, table = self.tables.find_le(chunk_id)

            # This should be a ewf_table_entry object but the below is faster.
            try:
                table_entry = table[chunk_id - start_chunk]

                offset = table_entry & 0x7fffffff
                next_offset = table[chunk_id - start_chunk + 1] & 0x7fffffff
                compressed_chunk_size = next_offset - offset
            except IndexError:
                return ""

            data = self.address_space.read(
                offset + table_header.base_offset, compressed_chunk_size)

            if table_entry & 0x80000000:
                data = zlib.decompress(data)

            # Cache the chunk for later.
            self.chunk_cache.Put(chunk_id, data)

            return data

    def read_partial(self, offset, length):
        """Read as much as possible from the current offset."""
        # Find the table responsible for this chunk.
        chunk_id, chunk_offset = divmod(offset, self.chunk_size)
        available_length = min(length, self.chunk_size - chunk_offset)

        # Get the chunk and split it.
        data = self.read_chunk(chunk_id)

        return data[chunk_offset:chunk_offset + available_length]

    def read(self, offset, length):
        """Read data from the file."""
        # Most read operations are very short and will not need to merge chunks
        # at all. In that case concatenating strings is much faster than storing
        # partial reads into a list and join()ing them.
        result = ''
        available_length = length

        while available_length > 0:
            buf = self.read_partial(offset, available_length)
            if not buf:
                break

            result += buf
            offset += len(buf)
            available_length -= len(buf)

        return result


class EWFFileWriter(object):
    """A writer for EWF files.

    NOTE: The EWF files we produce here are not generally compatible with
    Encase/FTK. We produce EWFv1 files which are unable to store sparse
    images. We place an ELF file inside the EWF container to ensure we can
    efficiently store sparse memory ranges.
    """

    def __init__(self, out_as, session):
        self.out_as = out_as
        self.session = session
        self.profile = EWFProfile(session=self.session)
        self.chunk_size = 32 * 1024
        self.current_offset = 0
        self.chunk_id = 0

        self.last_section = None

        # Start off by writing the file header.
        file_header = self.profile.ewf_file_header_v1(
            offset=0, vm=out_as)
        file_header.EVF_sig = file_header.EVF_sig.signature
        file_header.fields_start = 1
        file_header.segment_number = 1
        file_header.fields_end = 1

        self.current_offset = file_header.obj_end
        self.buffer = ""
        self.table_count = 0

        # Get ready to accept data.
        self.StartNewTable()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, trace):
        self.Close()

    def AddNewSection(self, section):
        if self.last_section:
            self.last_section.next = section
            self.last_section.UpdateChecksum()

        self.last_section = section

    def StartNewTable(self):
        """Writes a sections table and begins collecting chunks into table."""
        self.table = []
        self.table_count += 1

        sectors_section = self.profile.ewf_section_descriptor_v1(
            offset=self.current_offset, vm=self.out_as)
        sectors_section.type = "sectors"
        self.AddNewSection(sectors_section)

        self.base_offset = self.current_offset = sectors_section.obj_end

    def write(self, data):
        """Writes the data into the file.

        This method allows the writer to be used as a file-like object.
        """
        self.buffer += data
        buffer_offset = 0
        while len(self.buffer) - buffer_offset >= self.chunk_size:
            data = self.buffer[buffer_offset:buffer_offset+self.chunk_size]
            cdata = zlib.compress(data)
            chunk_offset = self.current_offset - self.base_offset

            if len(cdata) > len(data):
                self.table.append(chunk_offset)
                cdata = data
            else:
                self.table.append(0x80000000 | chunk_offset)

            self.out_as.write(self.current_offset, cdata)
            self.current_offset += len(cdata)
            buffer_offset += self.chunk_size
            self.chunk_id += 1

            # Flush the table when it gets too large. Tables can only store 31
            # bit offset and so can only address roughly 2gb. We choose to stay
            # under 1gb: 30000 * 32kb = 0.91gb.
            if len(self.table) > 30000:
                self.session.report_progress(
                    "Flushing EWF Table %s.", self.table_count)
                self.FlushTable()
                self.StartNewTable()

        self.buffer = self.buffer[buffer_offset:]

    def FlushTable(self):
        """Flush the current table."""
        table_section = self.profile.ewf_section_descriptor_v1(
            offset=self.current_offset, vm=self.out_as)
        table_section.type = "table"
        self.AddNewSection(table_section)

        table_header = self.profile.ewf_table_header_v1(
            offset=table_section.obj_end, vm=self.out_as)

        table_header.number_of_entries = len(self.table)
        table_header.base_offset = self.base_offset
        table_header.UpdateChecksum()

        # Now write the table section.
        self.out_as.write(
            table_header.entries.obj_offset,
            struct.pack("I" * len(self.table), *self.table))

        self.current_offset = (table_header.entries.obj_offset +
                               4 * len(self.table))
    def Close(self):
        # If there is some data left over, pad it to the length of the chunk so
        # we get to write it.
        if len(self.buffer):
            self.write("\x00" * (self.chunk_size - len(self.buffer)))

        self.FlushTable()

        # Write the volume section.
        volume_section = self.profile.ewf_section_descriptor_v1(
            offset=self.current_offset, vm=self.out_as)
        volume_section.type = "volume"
        self.AddNewSection(volume_section)

        volume_header = self.profile.ewf_volume(
            offset=volume_section.obj_end, vm=self.out_as)

        volume_header.number_of_chunks = self.chunk_id
        volume_header.sectors_per_chunk = self.chunk_size / 512
        volume_header.number_of_sectors = (volume_header.number_of_chunks *
                                           volume_header.sectors_per_chunk)

        volume_header.bytes_per_sector = 512
        volume_header.UpdateChecksum()

        # Write the done section.
        done_section = self.profile.ewf_section_descriptor_v1(
            offset=volume_header.obj_end, vm=self.out_as)
        done_section.type = "done"
        self.AddNewSection(done_section)

        # Last section points to itself.
        self.AddNewSection(done_section)


class EWFAcquire(plugin.PhysicalASMixin, plugin.Command):
    """Copy the physical address space to an EWF file."""

    name = "ewfacquire"

    @classmethod
    def args(cls, parser):
        super(EWFAcquire, cls).args(parser)

        parser.add_argument(
            "destination", default=None, required=False,
            help="The destination file to create. "
            "If not specified we write output.E01 in current directory.")

    def __init__(self, destination=None, **kwargs):
        super(EWFAcquire, self).__init__(**kwargs)

        self.destination = destination

    def render(self, renderer):
        if self.destination is None:
            out_fd = renderer.open(filename="output.E01", mode="w+b")
        else:
            directory, filename = os.path.split(self.destination)
            out_fd = renderer.open(filename=filename, directory=directory,
                                   mode="w+b")

        with out_fd:
            runs = list(self.physical_address_space.get_available_addresses())

            out_address_space = standard.WritableFDAddressSpace(
                fhandle=out_fd, session=self.session)

            with EWFFileWriter(
                out_address_space, session=self.session) as writer:
                if len(runs) > 1:
                    elfcore.WriteElfFile(
                        self.physical_address_space,
                        writer, session=self.session)

                else:
                    last_address = runs[0][0] + runs[0][2]
                    block_size = 1024 * 1024

                    for offset in xrange(0, last_address, block_size):
                        available_length = min(block_size, last_address-offset)
                        data = self.physical_address_space.read(
                            offset, available_length)

                        self.session.report_progress(
                            "Writing %sMB", offset/1024/1024)

                        writer.write(data)


class TestEWFAcquire(testlib.HashChecker):
    PARAMETERS = dict(commandline="ewfacquire %(tempdir)s/output_image.e01")



# Rekall Memory Forensics
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

__author__ = "Michael Cohen <scudette@gmail.com>"

from rekall import obj
from rekall import utils

from rekall.plugins.addrspaces import amd64
from rekall.plugins.overlays import basic


darwin_overlay = {
    "proc": [None, {
        # Some standard fields for Darwin processes.
        "name": lambda x: x.p_comm,
        "pid": lambda x: x.p_pid.v(),
        "dtb": lambda x: x.task.map.pmap.pm_cr3.v(),

        "p_list": [None, ["LIST_ENTRY"]],
        "p_sibling": [None, ["LIST_ENTRY"]],

        "p_comm": [None, ["UnicodeString", dict(length=17)]],
        "task": [None, ["Pointer", dict(
            target="task"
        )]],
    }],

    "task": [None, {
        "bsd_info": [None, ["Pointer", dict(target="proc")]],
    }],

    "rtentry": [None, {
        "base_calendartime": [None, ["UnixTimeStamp"]],

        "source_ip": lambda x: x.rt_nodes[0].rn_u.rn_leaf.m(
            "rn_Key").dereference_as("sockaddr"),

        "dest_ip": lambda x: x.rt_gateway.deref(),

        "name": lambda x: "%s%s" % (
            x.rt_ifp.if_name.dereference_as("String"),
            x.rt_ifp.if_unit),

        "sent": lambda x: x.m("rt_stats").nstat_txpackets,

        "rx": lambda x: x.rt_expire if x.m("rt_stats") else None,

        "delta": lambda x: (x.rt_expire - x.base_uptime
                            if x.rt_expire else 0),

    }],

    "kmod_info": [None, {
        "name": lambda x: utils.SmartUnicode(x.m("name").cast("UnicodeString")),
        "base": lambda x: x.address.v(),
        "end": lambda x: x.base + x.size,
        "version": [None, ["String"]],

        # Starting address of the kernel module.
        "address": [None, ["Pointer"]],
    }],

    "sockaddr": [None, {
        # github.com/opensource-apple/xnu/blob/10.9/bsd/sys/socket.h#L370
        "sa_family": [None, ["Enumeration", dict(
            enum_name="sa_family_t",
            target="unsigned char",
        )]],
    }],

    "sysctl_oid": [None, {
        # xnu-2422.1.72/bsd/sys/sysctl.h: 148
        # This field is reused for two purposes, the first is type of node,
        # while the second is the permissions.
        "oid_kind_type": lambda x: x.m("oid_kind").cast(
            "Enumeration", choices={
                1: "CTLTYPE_NODE",
                2: "CTLTYPE_INT",
                3: "CTLTYPE_STRING",
                4: "CTLTYPE_QUAD",
                5: "CTLTYPE_OPAQUE",
            },
            target="BitField",
            target_args=dict(start_bit=0, end_bit=8),
        ),

        "oid_perms": lambda x: x.m("oid_kind").cast(
            "Flags", maskmap={
                "CTLFLAG_RD": 0x80000000,  # Allow reads of variable */
                "CTLFLAG_WR": 0x40000000,  # Allow writes to the variable
                # A node will handle locking by itself.
                "CTLFLAG_LOCKED": 0x00800000,
            },
        ),

        "oid_name": [None, ["Pointer", dict(target="String")]],

    }],

    "zone": [None, {
        "zone_name": [None, ["Pointer", dict(target="String")]],
        "free_elements": [None, ["Pointer", dict(
            target="zone_free_element"
        )]],
    }],

    "vm_map_entry": [None, {
        # xnu-2422.1.72/osfmk/mach/vm_prot.h:81
        "protection": [None, ["Flags", dict(
            maskmap={
                "VM_PROT_READ": 1,
                "VM_PROT_WRITE": 2,
                "VM_PROT_EXECUTE": 4,
            },
            target="BitField",
            target_args=dict(
                start_bit=7,
                end_bit=10
            )
        )]],
        "max_protection": [None, ["Flags", dict(
            maskmap={
                "VM_PROT_READ": 1,
                "VM_PROT_WRITE": 2,
                "VM_PROT_EXECUTE": 4,
            },
            target="BitField",
            target_args=dict(
                start_bit=10,
                end_bit=13
            )
        )]],
    }],

    "vnode": [None, {
        "v_name": [None, ["Pointer", dict(target="String")]],

        "path": lambda self: "/".join(reversed(
            [unicode(y.v_name.deref()) for y in self.walk_list("v_parent")])),

        # xnu-2422.1.72/bsd/sys/vnode_internal.h:230
        "v_flag": [None, ["Flags", dict(
            maskmap={
                "VROOT": 0x000001,
                "VTEXT": 0x000002,
                "VSYSTEM": 0x000004,
                "VISTTY": 0x000008,
                "VRAGE": 0x000010,
            }
        )]],
    }],

    "cat_attr": [None, {
        "ca_atime": [None, ["UnixTimeStamp"]],
        "ca_atimeondisk": [None, ["UnixTimeStamp"]],
        "ca_mtime": [None, ["UnixTimeStamp"]],
        "ca_ctime": [None, ["UnixTimeStamp"]],
        "ca_itime": [None, ["UnixTimeStamp"]],
        "ca_btime": [None, ["UnixTimeStamp"]],
    }],

    "ifnet": [None, {
        "if_name": [None, ["Pointer", dict(target="String")]],
    }],

    "session": [None, {
        "s_login": [None, ["String"]],
    }],

    "filedesc": [None, {
        # Defined here:
        # https://github.com/opensource-apple/xnu/blob/10.9/bsd/sys/filedesc.h#L113
        "fd_ofileflags": [None, ["Pointer", dict(
            target="Array",
            target_args=dict(
                target="Flags",
                target_args=dict(
                    target="unsigned char",
                    maskmap=utils.MaskMapFromDefines("""
/*
 * Per-process open flags.
 */
#define UF_EXCLOSE      0x01            /* auto-close on exec */
#define UF_FORKCLOSE    0x02            /* auto-close on fork */
#define UF_RESERVED     0x04            /* open pending / in progress */
#define UF_CLOSING      0x08            /* close in progress */

#define UF_RESVWAIT     0x10            /* close in progress */
#define UF_INHERIT      0x20            /* "inherit-on-exec" */
"""))))]],

        "fd_ofiles": [None, ["Pointer", dict(
            target="Array",
            target_args=dict(
                target="Pointer",
                count=lambda x: x.fd_lastfile,
                target_args=dict(
                    target="fileproc"
                )
            )
        )]],
    }],

    "mount": [None, {
        # xnu-2422.1.72/bsd/sys/mount.h
        "mnt_flag": [None, ["Flags", dict(
            maskmap={
                "MNT_LOCAL": 0x00001000,
                "MNT_QUOTA": 0x00002000,
                "MNT_ROOTFS": 0x00004000,
            }
        )]],
    }],

    "vfsstatfs": [None, {
        "f_mntonname": [None, ["String"]],
        "f_mntfromname": [None, ["String"]],
        "f_fstypename": [None, ["String"]],
    }],

    "sockaddr_un": [None, {
        # sun_len is the number of bytes from start of the struct to the NULL
        # terminator in the sun_path member. (Yes, really. [1]) The
        # original/Volatility code used it as length of the string itself,
        # leading to a subtle bug, when the sun_path wasn't NULL-terminated
        # (which it doesn't have to be.)
        # [1]
        # https://github.com/opensource-apple/xnu/blob/10.9/bsd/sys/un.h#L77
        "sun_path": [None, ["String", dict(
            length=lambda x: (
                x.sun_len
                - x.obj_profile.get_obj_offset("sockaddr_un", "sun_path")
            ),
        )]],
    }],

    "domain": [None, {
        # xnu-2422.1.72/bsd/sys/domain.h: 99
        "dom_family": [None, ["Enumeration", dict(
            enum_name="sa_family_t",
            target="unsigned char",
        )]],
    }],

    "tcpcb": [None, {
        "t_state": [None, ["Enumeration", dict(
            # xnu-2422.1.72/bsd/netinet/tcp_fsm.h: 75
            choices={
                0: "TCPS_CLOSED",
                1: "TCPS_LISTEN",
                2: "TCPS_SYN_SENT",
                3: "TCPS_SYN_RECEIVED",
                4: "TCPS_ESTABLISHED",
                5: "TCPS_CLOSE_WAIT",
                6: "TCPS_FIN_WAIT_1",
                7: "TCPS_CLOSING",
                8: "TCPS_LAST_ACK",
                9: "TCPS_FIN_WAIT_2",
                10: "TCPS_TIME_WAIT",
            },
            target="int",
        )]],
    }],

    "protosw": [None, {
        "pr_protocol": [None, ["Enumeration", dict(
            # xnu-2422.1.72/bsd/netinet/in.h: 99
            choices={
                0: "IPPROTO_IP",
                1: "IPPROTO_ICMP",
                2: "IPPROTO_IGMP",
                4: "IPPROTO_IPV4",
                6: "IPPROTO_TCP",
                17: "IPPROTO_UDP",
                41: "IPPROTO_IPV6",
                50: "IPPROTO_ESP",
                51: "IPPROTO_AH",
                58: "IPPROTO_ICMPV6",
                255: "IPPROTO_RAW",
            },
            target="short",
        )]],
        "pr_type": [None, ["Enumeration", dict(
            enum_name="pr_type",
            target="short",
        )]],
    }],

    "OSDictionary": [None, {
        "dictionary": [None, ["Pointer", dict(
            target="Array",
            target_args=dict(
                target="dictEntry",
                count=lambda x: x.count,
            )
        )]],
    }],

    "OSString": [None, {
        "value": lambda x: x.obj_profile.UnicodeString(
            offset=x.string,
            length=x.length
        ),
    }],

    "OSOrderedSet": [None, {
        "array": [None, ["Pointer", dict(
            target="Array",
            target_args=dict(
                target="_Element",
                count=lambda x: x.count
            )
        )]],
    }],

    "_Element": [8, {
        "obj": [0, ["pointer", ["OSMetaClassBase"]]]
    }],

    "PE_state": [None, {
        "bootArgs": [None, ["Pointer", dict(target="boot_args")]],
    }],

    "EfiMemoryRange": [None, {
        # xnu-1699.26.8/pexpert/pexpert/i386/boot.h: 46
        "Type": [None, ["Enumeration", dict(
            choices={
                0: "kEfiReservedMemoryType",
                1: "kEfiLoaderCode",
                2: "kEfiLoaderData",
                3: "kEfiBootServicesCode",
                4: "kEfiBootServicesData",
                5: "kEfiRuntimeServicesCode",
                6: "kEfiRuntimeServicesData",
                7: "kEfiConventionalMemory",
                8: "kEfiUnusableMemory",
                9: "kEfiACPIReclaimMemory",
                10: "kEfiACPIMemoryNVS",
                11: "kEfiMemoryMappedIO",
                12: "kEfiMemoryMappedIOPortSpace",
                13: "kEfiPalCode",
                14: "kEfiMaxMemoryType",
            },
            target="unsigned int"
        )]],
    }]
}


darwin_enums = {
    # Socket address families, defined here because we use them in a couple of
    # places. This is a "fake enum" in the sense that the values are declared
    # as #defines, but only ever used in conjunction with a specific datatype
    # (which is sa_family_t).
    #
    # From the horse's mouth:
    # https://github.com/opensource-apple/xnu/blob/10.9/bsd/sys/socket.h#L370
    #
    # The data type is defined here:
    # https://github.com/opensource-apple/xnu/blob/10.9/bsd/sys/_types/_sa_family_t.h?source=cc#L30
    #
    # Basically the exact same fake enumeration is used in the same file for
    # protocol families here:
    # https://github.com/opensource-apple/xnu/blob/10.9/bsd/sys/socket.h#L493
    #
    # Because the numbers are the same (in fact, PF_ defines are mapped 1:1
    # with addressing families) we just use AF_ for everything.
    "sa_family_t": utils.EnumerationFromDefines("""
/*
 * Address families.
 */
#define AF_UNSPEC       0               /* unspecified */
#define AF_UNIX         1               /* local to host (pipes) */
#define AF_INET         2               /* internetwork: UDP, TCP, etc. */
#define AF_IMPLINK      3               /* arpanet imp addresses */
#define AF_PUP          4               /* pup protocols: e.g. BSP */
#define AF_CHAOS        5               /* mit CHAOS protocols */
#define AF_NS           6               /* XEROX NS protocols */
#define AF_ISO          7               /* ISO protocols */
#define AF_ECMA         8               /* European computer manufacturers */
#define AF_DATAKIT      9               /* datakit protocols */
#define AF_CCITT        10              /* CCITT protocols, X.25 etc */
#define AF_SNA          11              /* IBM SNA */
#define AF_DECnet       12              /* DECnet */
#define AF_DLI          13              /* DEC Direct data link interface */
#define AF_LAT          14              /* LAT */
#define AF_HYLINK       15              /* NSC Hyperchannel */
#define AF_APPLETALK    16              /* Apple Talk */
#define AF_ROUTE        17              /* Internal Routing Protocol */
#define AF_LINK         18              /* Link layer interface */
#define pseudo_AF_XTP   19              /* eXpress Transfer Protocol (no AF) */
#define AF_COIP         20              /* connection-oriented IP, aka ST II */
#define AF_CNT          21              /* Computer Network Technology */
#define pseudo_AF_RTIP  22              /* Help Identify RTIP packets */
#define AF_IPX          23              /* Novell Internet Protocol */
#define AF_SIP          24              /* Simple Internet Protocol */
#define pseudo_AF_PIP   25              /* Help Identify PIP packets */
#define AF_NDRV         27              /* Network Driver 'raw' access */
#define AF_ISDN         28              /* Integrated Services Digital Network*/
#define pseudo_AF_KEY   29              /* Internal key-management function */
#define AF_INET6        30              /* IPv6 */
#define AF_NATM         31              /* native ATM access */
#define AF_SYSTEM       32              /* Kernel event messages */
#define AF_NETBIOS      33              /* NetBIOS */
#define AF_PPP          34              /* PPP communication protocol */
#define pseudo_AF_HDRCMPLT 35           /* Used by BPF to not rewrite headers
                                         * in interface output routine */
#define AF_AFP          36              /* Used by AFP */
#define AF_IEEE80211    37              /* IEEE 802.11 protocol */
#define AF_UTUN         38
#define AF_MULTIPATH    39
#define AF_MAX          40
"""),

    # Socket types, defined here:
    # https://github.com/opensource-apple/xnu/blob/10.9/bsd/sys/socket.h#L133
    "pr_type": utils.EnumerationFromDefines("""
/*
 * Types
 */
#define SOCK_STREAM     1               /* stream socket */
#define SOCK_DGRAM      2               /* datagram socket */
#define SOCK_RAW        3               /* raw-protocol interface */
#define SOCK_RDM        4               /* reliably-delivered message */
#define SOCK_SEQPACKET  5               /* sequenced packet stream */
"""),
}

darwin64_types = {
    "LIST_ENTRY": [16, {
        "le_next": [0, ["Pointer"]],
        "le_prev": [8, ["Pointer", dict(target="Pointer")]],
    }],
}


class LIST_ENTRY(obj.Struct):
    """XNU defines lists inline using an annonymous struct. This makes it hard
    for us to automatically support lists because the debugging symbols dont
    indicate this inner struct is of any particular type (since its annonymous).

    We therefore depend on the overlays to redefine each list memeber as a
    LIST_ENTRY member. For example we see code like:

    struct proc {
       LIST_ENTRY(proc) p_list;
       ...

    Where:

    #define  LIST_ENTRY(type)                                       \
      struct {                                                      \
      struct type *le_next;  /* next element */                     \
      struct type **le_prev; /* address of previous next element */ \
    }
    """
    _forward = "le_next"
    _backward = "le_prev"

    def is_valid(self):
        """Must have both valid next and prev pointers."""
        return (self.m(self._forward).is_valid() and
                self.m(self._backward).is_valid())

    def _GetNextEntry(self, type, member):
        return self.m(self._forward).dereference_as(type).m(member)

    def _GetPreviousEntry(self):
        return self.m(self._backward).dereference_as(self.obj_type)

    def dereference_as(self, type, member, vm=None):
        """Recasts the list entry as a member in a type, and return the type.

        Args:
        type: The name of this Struct type.
        member: The name of the member of this Struct.
        address_space: An optional address space to switch during
        deferencing.
        """
        offset = self.obj_profile.get_obj_offset(type, member)

        item = self.obj_profile.Object(
            type_name=type, offset=self.obj_offset - offset,
            vm=vm or self.obj_vm, parent=self.obj_parent,
            name=type, context=self.obj_context)

        return item

    def find_all_lists(self, type, member, seen=None):
        """Follows all the list entries starting from lst.

        We basically convert the list to a tree and recursively search it for
        new nodes. From each node we follow the Flink and then the Blink. When
        we see a node we already have, we backtrack.
        """
        if seen is None:
            seen = set()

        if not self.is_valid():
            return seen

        elif self in seen:
            return seen

        seen.add(self)

        flink = self._GetNextEntry(type, member)
        flink.find_all_lists(type, member, seen=seen)

        blink = self._GetPreviousEntry()
        blink.find_all_lists(type, member, seen=seen)

        return seen

    def list_of_type(self, type, member=None, include_current=True):
        # We sort here to ensure we have stable ordering as the output of this
        # call.
        result = sorted(self.find_all_lists(type, member),
                        key=lambda x: x.obj_offset)

        if member is None:
            member = self.obj_name

        # Return ourselves as the first item.
        if include_current:
            yield self.dereference_as(type, member)

        # We traverse all the _LIST_ENTRYs we can find, and cast them all back
        # to the required member.
        for lst in result:
            # Skip ourselves in this (list_of_type is usually invoked on a list
            # head).
            if lst.obj_offset == self.obj_offset:
                continue

            task = lst.dereference_as(type, member)
            if task:
                # Only yield valid objects (In case of dangling links).
                yield task

    def reflect(self, vm=None):
        """Reflect this list element by following its Flink and Blink.

        This is basically the same as Flink.Blink except that it also checks
        Blink.Flink. It also ensures that Flink and Blink are dereferences to
        the correct type in case the vtypes do not specify them as pointers.

        Returns:
        the result of Flink.Blink.
        """
        result1 = self.m(self._forward).dereference_as(
            self.obj_type, vm=vm).m(self._backward).deref().cast(
                self.obj_type)

        if not result1:
            return obj.NoneObject("Flink not valid.")

        result2 = self.Blink.deref().dereference_as(
            self.obj_type, vm=vm).m(
                self._forward).dereference_as(self.obj_type)

        if result1 != result2:
            return obj.NoneObject("Flink and Blink not consistent.")

        return result1

    def __nonzero__(self):
        # List entries are valid when both Flinks and Blink are valid
        return bool(self.m(self._forward)) or bool(self.m(self._backward))

    def __iter__(self):
        return self.list_of_type(self.obj_parent.obj_type, self.obj_name)


class queue_entry(LIST_ENTRY):
    _forward = "next"
    _backward = "prev"


class sockaddr_dl(obj.Struct):
    def __unicode__(self):
        result = []
        for i in xrange(self.sdl_alen):
            result.append(
                "%.02X" % ord(self.sdl_data[self.sdl_nlen + i].v()))

        return ":".join(result)


class fileproc(obj.Struct):
    """Represents an open file, owned by a process."""

    DTYPE_TO_HUMAN = {
        "DTYPE_SOCKET": "socket",
        "DTYPE_VNODE": "vnode",
        "DTYPE_PSXSEM": "POSIX Semaphore",
        "DTYPE_PSXSHM": "POSIX Shared Mem.",
        "DTYPE_KQUEUE": "kernel queue",
        "DTYPE_PIPE": "pipe",
        "DTYPE_FSEVENTS": "FS Events",  # needs more research

        # I (Adam) /believe/ this is a remnant of the AppleTalk support,
        # however nowadays it's unfortunately often used to mean basically
        # DTYPE_OTHER (which XNU doesn't have). Example of how this might be
        # used in current code:
        # opensource.apple.com/source/xnu/xnu-1456.1.26/bsd/netat/sys_dep.c
        "DTYPE_ATALK": "<unknown>",
        "-": "INVALID",
    }

    @property
    def fg_type(self):
        """Returns type of the fileglob (e.g. vnode, socket, etc.)"""
        return self.multi_m(
            # OS X 10.8 and earlier
            "f_fglob.fg_type",

            # OS X 10.9 and later
            "f_fglob.fg_ops.fo_type")

    def autocast_fg_data(self):
        """Returns the correct struct with fg_type-specific information.

        This can be one of vnode, socket, shared memory or semaphore [1].

        Of those four, we currently only get extra information for vnode and
        socket. For everything else, we return a NoneObject.

        [1]:
          https://github.com/opensource-apple/xnu/blob/10.9/bsd/sys/file_internal.h#L184
        """
        dtype = self.fg_type

        # Semaphore and shared memory are known structs, but we currently don't
        # know of anything interesting that should be extracted from them.
        if dtype == "DTYPE_SOCKET":
            return self.f_fglob.fg_data.dereference_as("socket")
        elif dtype == "DTYPE_VNODE":
            return self.f_fglob.fg_data.dereference_as("vnode")
        # elif dtype == "DTYPE_PSXSEM":
        #     return self.f_fglob.fg_data.dereference_as("semaphore")
        # elif dtype == "DTYPE_PSXSHM":
        #     return self.f_fglob.fg_data.dereference_as("vm_shared_region")

        # That would be an unknown DTYPE.
        return self.f_fglob.fg_data

    @property
    def human_name(self):
        return getattr(self.autocast_fg_data(), "human_name", None)

    @property
    def human_type(self):
        # Delegate to fg_data if it thinks it knows what it is.
        return getattr(
            self.autocast_fg_data(),
            "human_type",
            self.DTYPE_TO_HUMAN[str(self.fg_type)]
        )


class socket(obj.Struct):
    """Provides human-readable accessors for sockets of the more common AFs.

    This class has two basic ways of getting information. Most attributes are
    computed using the method fill_socketinfo, which is directly adapted from
    the kernel function of the same name. For the few things that
    fill_socketinfo doesn't care about, the properties themselves get the
    data and provide references to the kernel source for anyone wondering
    why and how all this works.
    """

    cached_socketinfo = None

    def fill_socketinfo(self):
        """Computes information about sockets of some addressing families.

        This function is directly adapted from the kernel function
        fill_socketinfo [1]. The original function is used to fill a struct
        with addressing and other useful information about sockets of a few
        key addressing families. All families are supported, but only the
        following will return useful information:
          - AF_INET (IPv4)
          - AF_INET6 (IPv6)
          - AF_UNIX (Unix socket)
          - AF_NDRV (Network driver raw access)
          - AF_SYSTEM (Darwin-specific; see documentation [3])

        Differences between the kernel function and this adaptation:
          - The kernel uses Protocol Families (prefixed with PF_). Rekall
          relies on Addressing Families (AF_) which are exactly the same.

          - The kernel fills a struct; this function returns a dict with the
          same members.

          - The kernel returns the data raw. This function converts endianness
          and unions to human-readable representations, as appropriate.

          - Only a subset of members are filled in.

          - Other differences as documented in code.

        Returns:
          A dict with the same members as struct socket_info and related.
          Only member that's always filled is "soi_kind". That's not Spanish,
          but one of the values in this anonymous enum [2], which determines
          what other members are present. (Read the code.)

        [1]
        https://github.com/opensource-apple/xnu/blob/10.9/bsd/kern/socket_info.c#L98
        [2]
        https://github.com/opensource-apple/xnu/blob/10.9/bsd/sys/proc_info.h#L503
        [3] "KEXT Controls and Notifications"
        https://developer.apple.com/library/mac/documentation/Darwin/Conceptual/NKEConceptual/control/control.html
        """
        domain = self.so_proto.pr_domain.dom_family
        type = self.so_proto.pr_type
        protocol = self.so_proto.pr_protocol

        si = {"soi_kind": "SOCKINFO_GENERIC"}

        # The kind of socket is determined by the triplet
        # {domain, type, protocol}
        if domain in ["AF_INET", "AF_INET6"]:
            si["soi_kind"] = "SOCKINFO_IN"

            inp = self.so_pcb.dereference_as("inpcb")

            si["insi_fport"] = utils.ntoh(inp.inp_fport)
            si["insi_lport"] = utils.ntoh(inp.inp_lport)
            si["insi_ip_ttl"] = inp.inp_ip_ttl.v()

            # Different from kernel: insi_[df]addr is a union, and by setting
            # the IPv6 address, you set the IPv4 address too. We instead return
            # a string with the appropriately formatted address.
            if domain == "AF_INET":
                si["insi_faddr"] = utils.FormatIPAddress(
                    "AF_INET",
                    inp.inp_dependfaddr.inp46_foreign.ia46_addr4.s_addr
                )
                si["insi_laddr"] = utils.FormatIPAddress(
                    "AF_INET",
                    inp.inp_dependladdr.inp46_local.ia46_addr4.s_addr
                )
            else:
                si["insi_faddr"] = utils.FormatIPAddress(
                    "AF_INET6",
                    inp.inp_dependfaddr.inp6_foreign.m("__u6_addr")
                )
                si["insi_laddr"] = utils.FormatIPAddress(
                    "AF_INET6",
                    inp.inp_dependladdr.inp6_local.m("__u6_addr")
                )

            if (type == "SOCK_STREAM"
                    and (protocol == 0 or protocol == "IPPROTO_TCP")
                    and inp.inp_ppcb != None):

                tp = inp.inp_ppcb.dereference_as("tcpcb")
                si["soi_kind"] = "SOCKINFO_TCP"

                si["tcpsi_state"] = tp.t_state
                si["tcpsi_flags"] = tp.t_flags
        elif domain == "AF_UNIX":
            unp = self.so_pcb.dereference_as("unpcb")
            si["soi_kind"] = "SOCKINFO_UN"

            if unp.unp_addr:
                # Difference from kernel: instead of copying the whole unp_addr
                # struct, we just get delegate getting the actual string to the
                # unp_addr struct. (Because it's trickier than it looks.)
                si["unsi_addr"] = unp.unp_addr.sun_path

        elif domain == "AF_NDRV":
            # This is how we get the pcb if we need to:
            # ndrv_cb = self.so_pcb.dereference_as("ndrv_cb")
            si["soi_kind"] = "SOCKINFO_NDRV"
        elif domain == "AF_SYSTEM":
            # AF_SYSTEM domain needs more research. It looks like it's used to
            # communicate between user space and kernel extensions, and allows
            # the former to control the latter. Naively, this looks ripe for
            # rootkits to me.
            if protocol == "SYSPROTO_EVENT":
                # This is how we get the pcb if we need to:
                # ev_pcb = self.so_pcb.dereference_as("kern_event_pcb")
                si["soi_kind"] = "SOCKINFO_KERN_EVENT"
            elif protocol == "SYSPROTO_CONTROL":
                kcb = self.so_pcb.dereference_as("ctl_cb")
                kctl = kcb.kctl
                si["soi_kind"] = "SOCKINFO_KERN_CTL"

                if kctl:
                    si["kcsi_id"] = kctl.id
                    si["kcsi_name"] = kctl.name

        return si

    def get_socketinfo_attr(self, attr):
        """Run fill_socketinfo if needed, cache result, return value of attr."""
        if not self.cached_socketinfo:
            self.cached_socketinfo = self.fill_socketinfo()

        if attr not in self.cached_socketinfo:
            return obj.NoneObject(
                "socket of family {}/{} has no member {}".format(
                    self.addressing_family,
                    self.cached_socketinfo["soi_kind"],
                    attr))

        return self.cached_socketinfo[attr]

    @property
    def src_addr(self):
        """For IPv[46] sockets, return source IP as string."""
        return self.get_socketinfo_attr("insi_laddr")

    @property
    def dst_addr(self):
        """For IPv[46] sockets, return destination IP as string."""
        return self.get_socketinfo_attr("insi_faddr")

    @property
    def addressing_family(self):
        """The Addressing Family corresponds roughly to OSI layer 3."""
        return self.so_proto.pr_domain.dom_family

    @property
    def tcp_state(self):
        return self.get_socketinfo_attr("tcpsi_state")

    @property
    def vnode(self):
        """For Unix sockets, pointer to vnode, if any.

        This is the same way that OS gathers this information in response to
        syscall [1] (this is the API used by netstat, among others).

        1:
        https://github.com/opensource-apple/xnu/blob/10.9/bsd/kern/uipc_usrreq.c#L1683
        """
        if self.addressing_family == "AF_UNIX":
            return self.so_pcb.dereference_as("unpcb").unp_vnode

    @property
    def unp_conn(self):
        """For Unix sockets, the pcb of the paired socket. [1]

        You most likely want to do sock.conn_pcb.unp_socket to get at the
        other socket in the pair. However, because the sockets are paired
        through the protocol control block, it's actually useful to have
        a direct pointer at it in order to be able to spot paired sockets.

        1:
        https://github.com/opensource-apple/xnu/blob/10.9/bsd/sys/unpcb.h#L128
        """
        if self.addressing_family == "AF_UNIX":
            return self.so_pcb.dereference_as("unpcb").unp_conn

    @property
    def src_port(self):
        return self.get_socketinfo_attr("insi_lport")

    @property
    def dst_port(self):
        return self.get_socketinfo_attr("insi_fport")

    @property
    def l4_protocol(self):
        if self.addressing_family in ["AF_INET", "AF_INET6"]:
            # All the values start with IPPROTO_.
            return str(self.so_proto.pr_protocol).replace("IPPROTO_", "")

    @property
    def unix_type(self):
        if self.addressing_family == "AF_UNIX":
            pr_type = str(self.so_proto.pr_type)

            if pr_type:
                # All values begin with SOCK_.
                return pr_type.replace("SOCK_", "")
            else:
                # I am about 80% sure that this should never happen. Before
                # deciding how this should be handled (possibly by logging an
                # error), I'll need to do more research.
                return "Unix Socket"

    @property
    def human_name(self):
        if self.addressing_family in ["AF_INET", "AF_INET6"]:
            if self.l4_protocol in ["TCP", "UDP"]:
                return "{} ({}) -> {} ({})".format(
                    self.src_addr, self.src_port,
                    self.dst_addr, self.dst_port)

            return "{} -> {}".format(self.src_addr, self.dst_addr)

        if self.addressing_family == "AF_UNIX":
            return self.get_socketinfo_attr("unsi_addr")

        return None

    @property
    def human_type(self):
        if self.addressing_family == "AF_INET":
            return "{}v4".format(self.l4_protocol)

        if self.addressing_family == "AF_INET6":
            proto = self.l4_protocol

            # Some v6 protocols are already named with v6 in the name.
            if proto.endswith("6"):
                return proto

            return "{}v6".format(self.l4_protocol)

        if self.addressing_family == "AF_UNIX":
            return self.unix_type

        return "Sock: {}".format(self.addressing_family)


class sockaddr(obj.Struct):
    def _get_address_obj(self):
        addr = obj.NoneObject("Unknown socket family")

        if self.sa_family == "AF_INET":
            addr = self.cast("sockaddr_in").sin_addr.s_addr

        elif self.sa_family == "AF_INET6":
            addr = self.cast("sockaddr_in6").sin6_addr.m("__u6_addr")

        elif self.sa_family == "AF_LINK":
            addr = self.cast("sockaddr_dl")

        return addr

    def __unicode__(self):
        result = ""
        addr = self._get_address_obj()
        if addr:
            if self.sa_family in ("AF_INET6", "AF_INET"):
                result = utils.FormatIPAddress(self.sa_family, addr)

            elif self.sa_family == "AF_LINK":
                result = addr

        return str(result)


class vm_map_entry(obj.Struct):
    def find_vnode_object(self):
        """Find the underlying vnode object for the given vm_map_entry.

        xnu-2422.1.72/osfmk/vm/bsd_vm.c: 1339.
        """
        if not self.is_sub_map.v():
            #/*
            #* The last object in the shadow chain has the
            #* relevant pager information.
            #*/

            shadow = self.last_shadow
            if not shadow:
                return shadow

            if (shadow and not shadow.internal.v() and
                    shadow.pager_ready.v() and
                    not shadow.terminating.v() and
                    shadow.alive.v()):
                memory_object = shadow.pager
                pager_ops = memory_object.mo_pager_ops

                # If this object points to the vnode_pager_ops, then we
                # found what we're looking for.  Otherwise, this
                # vm_map_entry doesn't have an underlying vnode and so we
                # fall through to the bottom and return NULL.

                if pager_ops == self.obj_profile.get_constant(
                        "_vnode_pager_ops", is_address=True):
                    return shadow.pager.dereference_as(
                        "vnode_pager").vnode_handle

        return obj.NoneObject("vnode not found")

    @property
    def sharing_mode(self):
        """Returns the sharing mode of the backing vm_object.

        This is losely adapted from vm_map.c, void vm_map_region_top_walk(),
        except we're not filling page counts for resident/reusable, etc.
        """
        if not self.object.vm_object or self.is_sub_map:
            return "SM_EMPTY"  # Nada.

        vmobj = self.object.vm_object
        ref_count = vmobj.ref_count
        if vmobj.paging_in_progress:
            ref_count -= 1

        if vmobj.shadow:
            return "SM_COW"  # Copy on write.

        if self.superpage_size:
            return "SM_LARGE_PAGE"  # Shared large (huge) page.

        if self.needs_copy:
            return "SM_COW"

        if ref_count == 1 or (not vmobj.pager_trusted and not
                              vmobj.internal):
            return "SM_PRIVATE"

        return "SM_SHARED"

    @property
    def code_signed(self):
        return self.last_shadow.code_signed

    @property
    def last_shadow(self):
        shadow = self.object.vm_object
        if not shadow:
            return obj.NoneObject("no vm_object found")

        while shadow.shadow:
            shadow = shadow.shadow

        return shadow


class clist(obj.Struct):

    @property
    def recovered_contents(self):
        """Gets the full contents of the ring buffer, which may be freed.

        This is different from getting the legal contents as with b_to_q [1]
        because clists are only used by TTYs and they seem to always be all
        marked as consumed, so b_to_q wouldn't let us see any content.

        1: github.com/opensource-apple/xnu/blob/10.9/bsd/kern/tty_subr.c#L358
        """
        return self.obj_vm.read(self.c_cs, self.c_cn)


class proc(obj.Struct):
    def get_open_files(self):
        """Gets all open files (sockets, pipes...) owned by this proc.

        Yields:
          tuple of (fd, fileproc, flags)
        """
        # lastfile is a high water mark of valid fds [1]. That doesn't mean
        # there are no invalid fds at lower indexes! fd_freefile is a free
        # descriptor that tends to gravitate towards the lowest index as
        # as seen here [2]. When the kernel frees an fd it sets the pointer
        # to NULL and also clears the corresponding index in fd_ofilesflags
        # [3]. This creates a sparse array, so the search has to skip over
        # invalid fds along the way, just as the kernel does [4]. We skip
        # NULL pointers (and invalid pointers) but don't check for cleared
        # flags, since they're usually zero anyway.
        #
        # [1]:
        # https://github.com/opensource-apple/xnu/blob/10.9/bsd/sys/filedesc.h#L96
        # [2]:
        # https://github.com/opensource-apple/xnu/blob/10.9/bsd/kern/kern_descrip.c#L412
        # [3]:
        # https://github.com/opensource-apple/xnu/blob/10.9/bsd/kern/kern_descrip.c#L384
        # [4]:
        # https://github.com/opensource-apple/xnu/blob/10.9/bsd/kern/kern_descrip.c#L2960
        last_fd = self.p_fd.fd_lastfile
        ofiles = self.p_fd.fd_ofiles.deref()
        ofileflags = self.p_fd.fd_ofileflags

        for fd in xrange(last_fd + 1):  # xrange stops at N-1.
            file_obj = ofiles[fd].deref()

            # file_obj will be None if the pointer is NULL (see ref [4]), and
            # also when the pointer is simply invalid, which can happen
            # sometimes. Currently, I chalk it up to inconsistencies in the
            # volatile RAM image (since it's rare) but it might have another
            # explanation.
            if file_obj:
                yield (fd, file_obj, ofileflags[fd])

    def get_process_address_space(self):
        cr3 = self.task.map.pmap.pm_cr3
        as_class = self.obj_vm.__class__
        if self.task.map.pmap.pm_task_map == "TASK_MAP_64BIT_SHARED":
            as_class = amd64.AMD64PagedMemory

        return as_class(base=self.obj_vm.base, session=self.obj_vm.session,
                        dtb=cr3, name="Pid %s" % self.p_pid)

    @property
    def argv(self):
        result = []
        array = self.obj_profile.ListArray(
            target="String",
            offset=self.user_stack - self.p_argslen,
            vm=self.get_process_address_space(),
            maximum_size=self.p_argslen,
        )

        for item in array:
            # Total size of the argv array is specified in argc (not counting
            # padding).
            if len(result) >= self.p_argc:
                break

            item = unicode(item)

            # The argv array may have null padding for alignment. Discard these
            # empty strings.
            if not len(item):
                continue

            result.append(item)

        # argv[0] is often repeated as the executable name, to avoid confusion,
        # we just discard it.
        if len(result) > 1 and result[0] == result[1]:
            result.pop(0)

        return result

    def validate(self):
        """Use heuristics to guess whether this proc is valid."""
        return (self.p_argc > 0
                and len(self.p_comm) > 0
                and self.p_start.v() > 0
                and 99999 > self.pid > 0)


class vnode(obj.Struct):
    @property
    def full_path(self):
        # TODO: Speed this up by caching the paths in the session.
        result = []
        _vnode = self

        # Iterate here until we hit the root of the filesystem.
        while not (_vnode.v_flag.VROOT and
                   _vnode.v_mount.mnt_flag.MNT_ROOTFS):
            result.append(_vnode.v_name.deref())

            # If there is no parent skip to the mount point.
            _vnode = _vnode.v_parent or _vnode.v_mount.mnt_vnodecovered

            # This is rare, but it does happen. I currently don't understand
            # why, so we just log a warning and report the node as an orphan.
            if not _vnode:
                self.obj_session.logging.warning("vnode at 0x%x is orphaned.",
                                                 int(_vnode))
                return "<Orphan>"

        path = "/" + "/".join((str(x) for x in reversed(result) if x))
        return unicode(path.encode("string-escape"))
        # return "/" + "/".join((unicode(x) for x in reversed(result) if x))

    @property
    def human_type(self):
        return "Reg. File"

    @property
    def human_name(self):
        return self.full_path


class OSDictionary(obj.Struct):
    """The OSDictionary is a general purpose associative array described:

    xnu-1699.26.8/libkern/libkern/c++/OSDictionary.h
    """

    def items(self, value_class=None):
        """Iterate over the associative array and yield key, value pairs."""
        for entry in self.dictionary:
            key = entry.key.dereference_as("OSString").value
            if value_class:
                yield key, entry.value.dereference_as(value_class)
            else:
                yield key, entry.value


class OSOrderedSet(obj.Struct):
    """An OSOrderedSet is a list of OSObject instances.

    xnu-1699.26.8/libkern/libkern/c++/OSOrderedSet.h
    """

    def list_of_type(self, type_name):
        for item in self.array:
            yield item.obj.dereference_as(type_name)


class Darwin32(basic.Profile32Bits, basic.BasicClasses):
    """A Darwin profile."""
    METADATA = dict(
        os="darwin",
        arch="I386",
        type="Kernel")

    @classmethod
    def Initialize(cls, profile):
        super(Darwin32, cls).Initialize(profile)

        # Some Darwin profiles add a suffix to IOKIT objects. So OSDictionary
        # becomes OSDictionary_class. We automatically generate the overlays and
        # classes to account for this.
        for k in profile.vtypes.keys():
            if k.endswith("_class"):
                stripped_k = k[:-len("_class")]
                if stripped_k not in profile.vtypes:
                    profile.vtypes[stripped_k] = profile.vtypes[k]
                    if stripped_k in darwin_overlay:
                        darwin_overlay[k] = darwin_overlay[stripped_k]

        profile.add_classes(
            LIST_ENTRY=LIST_ENTRY, queue_entry=queue_entry,
            sockaddr=sockaddr, sockaddr_dl=sockaddr_dl,
            vm_map_entry=vm_map_entry, proc=proc, vnode=vnode,
            socket=socket, clist=clist,
            # Support both forms with and without _class suffix.
            OSDictionary=OSDictionary, OSDictionary_class=OSDictionary,
            OSOrderedSet=OSOrderedSet, OSOrderedSet_class=OSOrderedSet,
            fileproc=fileproc)
        profile.add_enums(**darwin_enums)
        profile.add_overlay(darwin_overlay)
        profile.add_constants(default_text_encoding="utf8")

    def get_constant_cpp_object(self, constant, **kwargs):
        """A variant of get_constant_object which accounts for name mangling."""
        for key in self.constants:
            if constant in key:
                return self.get_constant_object(key, **kwargs)


class Darwin64(basic.RelativeOffsetMixin, basic.ProfileLP64, Darwin32):
    """Support for 64 bit darwin systems."""
    METADATA = dict(
        os="darwin",
        arch="AMD64",
        type="Kernel")

    image_base = 0

    @classmethod
    def Initialize(cls, profile):
        super(Darwin64, cls).Initialize(profile)
        profile.add_types(darwin64_types)

    def GetImageBase(self):
        if not self.image_base:
            self.image_base = self.session.GetParameter(
                "vm_kernel_slide", 0)

        return self.image_base



#!/usr/bin/python

# Rekall Memory Forensics
# Copyright (C) 2014 Michael Cohen <scudette@gmail.com>
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
"""This module discovers the kernel base address.

The profile provides kernel addresses relative to the kernel base address. This
varies each time, so we need a way to locate the kernel base address in the
kernel address space.
"""

__author__ = "Michael Cohen <scudette@google.com>"

from rekall import obj
from rekall import scan
from rekall.plugins.windows import common
from rekall.plugins.overlays.windows import pe_vtypes


class ExportScanner(scan.BaseScanner):
    # We search for the name of a section present in the PE header.
    checks = [("MultiStringFinderCheck", dict(needles=[
        "INITKDBG", "MISYSPTE", "PAGEKD"]))]


class ObjectTreeHook(common.AbstractWindowsParameterHook):
    """Cache the object tree."""

    name = "object_tree"

    def BuildTree(self, result, root):
        for x in root:
            name = x.NameInfo.Name.v()
            if name == None:
                continue

            # We store the _OBJECT_HEADER offset and some metadata about the
            # types.
            object_type = x.get_object_type()
            entry = result[name] = dict(
                type=object_type,
                type_name="_OBJECT_HEADER",
                offset=x.obj_offset,
            )

            if object_type == "Directory":
                children = entry["Children"] = {}
                self.BuildTree(children, x.Object)

    def calculate(self):
        root = self.session.profile.get_constant_object(
            "ObpRootDirectoryObject",
            target="Pointer",
            target_args=dict(
                target="_OBJECT_DIRECTORY"
                )
            )

        result = dict(
            type="Directory",
            type_name="_OBJECT_DIRECTORY",
            offset=root.deref().obj_offset,
            Children={})

        self.BuildTree(result["Children"], root)
        return result


class DriveLetterDeviceHook(common.AbstractWindowsParameterHook):
    """Maps device names to drive letters."""

    name = "drive_letter_device_map"

    def calculate(self):
        result = {}
        obj_tree_plugin = self.session.plugins.object_tree()
        # The global path contains symlinks from the drive letter to the device
        # name.
        for global_obj in obj_tree_plugin.GetObjectByName(r"\GLOBAL??").Object:
            name = global_obj.NameInfo.Name.v()
            if (global_obj.get_object_type() == "SymbolicLink" and
                    len(name) > 1 and name[1] == ":"):
                target = global_obj.Object.LinkTarget.v()

                result[target] = name

        return result


class KernelBaseHook(common.AbstractWindowsParameterHook):
    """Finds the kernel base address."""

    name = "kernel_base"

    def calculate(self):
        address_space = self.session.kernel_address_space
        if not address_space:
            return

        scanner = ExportScanner(session=self.session,
                                address_space=address_space)

        # The kernel image is always loaded in the same range called the
        # "Initial Loader Mappings". Narrowing the possible range makes scanning
        # much faster. (See http://www.codemachine.com/article_x64kvas.html)
        if self.session.profile.metadata("arch") == "AMD64":
            kernel_boundary = 0xFFFFF80000000000
        else:
            kernel_boundary = 0x80000000

        maxlen = 0xFFFFF87FFFFFFFFF - kernel_boundary
        kernel_boundary = obj.Pointer.integer_to_address(kernel_boundary)
        for hit in scanner.scan(offset=kernel_boundary, maxlen=maxlen):

            # Search backwards for an MZ signature on the page boundary.
            page = hit & 0xFFFFFFFFFFFFF000
            for _ in range(10):
                if address_space.read(page, 2) == "MZ":
                    helper = pe_vtypes.PE(
                        address_space=address_space,
                        session=self.session, image_base=page)

                    if str(helper.RSDS.Filename) in common.KERNEL_NAMES:
                        self.session.logging.info(
                            "Detected kernel base at 0x%X", page)
                        return page
                else:
                    page -= 0x1000


class WindowsHighestUserAddress(common.AbstractWindowsParameterHook):
    """The highest address for user mode/kernel mode division."""

    name = "highest_usermode_address"

    def calculate(self):
        return self.session.profile.get_constant_object(
            "MmHighestUserAddress", "Pointer").v()


class DTB2TaskMap(common.AbstractWindowsParameterHook):
    """Maps the DTB to the _EPROCESS structs."""

    name = "dtb2task"

    def calculate(self):
        result = {}
        for task in self.session.plugins.pslist().filter_processes():
            result[int(task.dtb)] = task.obj_offset

        return result



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
"""Plugins for inspecting memory allocation zones."""

__author__ = "Michael Cohen <scudette@google.com>"

from rekall import utils
from rekall.plugins.darwin import common


class DarwinDumpZone(common.DarwinPlugin):
    """Dumps an allocation zone's contents."""

    __name = "dump_zone"

    @classmethod
    def args(cls, parser):
        super(DarwinDumpZone, cls).args(parser)
        parser.add_argument("--zone", default="buf.512")

    def __init__(self, zone="buf.512", **kwargs):
        super(DarwinDumpZone, self).__init__(**kwargs)
        self.zone_name = zone

    def render(self, renderer):
        for entity in self.session.entities.find(
                query=("Buffer/purpose is 'zones' and "
                       "any Buffer/context matches "
                       " (AllocationZone/name is {zone_name})"),
                query_params=dict(zone_name=self.zone_name)):
            utils.WriteHexdump(
                renderer=renderer,
                data=entity["Buffer/contents"],
                base=entity["Buffer/address"][0])


class DarwinDeadProcesses(common.DarwinPlugin):
    """Show deallocated processes which still exist in the zone allocator."""

    __name = "dead_procs"

    def render(self, renderer):
        # Find the proc zone from the allocator.
        proc_zone = self.session.entities.find_first(
            "AllocationZone/name is 'proc'")["Struct/base"]

        # Walk over the free list and get all proc objects.
        procs = []
        for allocation in proc_zone.free_elements.walk_list("next"):
            proc = allocation.cast("proc")
            # Validate the proc.
            if proc.p_argc > 0:
                procs.append(proc)

        if procs:
            # Just delegate the rendering to the regular pslist plugin.
            pslist_plugin = self.session.plugins.pslist(proc=procs)
            pslist_plugin.render(renderer)



# Rekall Memory Forensics
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
The Rekall Memory Forensics entity layer.
"""
__author__ = "Adam Sindelar <adamsh@google.com>"

import datetime

from rekall.entities import definitions
from rekall.entities import collector


class BallastGenerator(collector.EntityCollector):
    """Generates ballast entities to stress-test the entity system."""

    outputs = ["Timestamps", "Named"]

    @classmethod
    def is_active(cls, session):
        return session.GetParameter("generate_ballast") > 0

    def collect(self, hint):
        for i in range(self.session.GetParameter("generate_ballast")):
            yield [
                definitions.Named(
                    name="Ballast entry #%d" % i,
                    kind="Ballast"),
                definitions.Timestamps(
                    created_at=datetime.datetime.fromtimestamp(0))]



# Rekall Memory Forensics
# Copyright (c) 2012, Michael Cohen <scudette@gmail.com>
# Copyright (c) 2010, 2011, 2012 Michael Ligh <michael.ligh@mnin.org>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

import yara

from rekall import plugin
from rekall import utils
from rekall.plugins import yarascanner
from rekall.plugins.linux import common



class LinYaraScan(common.LinProcessFilter):
    """Scan using yara signatures."""

    __name = "yarascan"

    @classmethod
    def args(cls, parser):
        super(LinYaraScan, cls).args(parser)
        parser.add_argument("--string", default=None,
                            help="A verbatim string to search for.")

        parser.add_argument("--yara_file", default=None,
                            help="The yara signature file to read.")

        parser.add_argument("--yara_expression", default=None,
                            help="If provided we scan for this yarra "
                            "expression.")

        parser.add_argument(
            "--scan_physical", default=False, type="Boolean",
            help="If specified we scan the physcial address space.")

    def __init__(self, string=None, scan_physical=False, yara_file=None,
                 yara_expression=None, **kwargs):
        """Scan using yara signatures.

        Args:
          string: A verbatim string to search for.
            we scan their entire address spaces.
          scan_physical: If true we scan the physical address space.
          yara_file: The yara file to read.
          yara_expression: If provided we scan for this yarra expression.
        """
        super(LinYaraScan, self).__init__(**kwargs)
        if yara_expression:
            self.rules_source = yara_expression
            self.rules = yara.compile(source=self.rules_source)

        elif string:
            self.rules_source = (
                'rule r1 {strings: $a = "%s" condition: $a}' % string
                )
            self.rules = yara.compile(source=self.rules_source)

        elif yara_file:
            self.rules = yara.compile(yara_file)
        else:
            raise plugin.PluginError("You must specify a yara rule file or "
                                     "string to match.")

        self.scan_physical = scan_physical

    def generate_hits(self, address_space, end=None):
        scanner = yarascanner.BaseYaraASScanner(
            profile=self.profile, session=self.session,
            address_space=address_space,
            rules=self.rules)

        return scanner.scan(maxlen=end)

    def render_scan_physical(self, renderer):
        """This method scans the process memory using the VAD."""
        for rule, address, _, _ in self.generate_hits(
                self.physical_address_space):
            renderer.format("Rule: {0}\n", rule)

            context = self.physical_address_space.read(address, 0x40)
            utils.WriteHexdump(renderer, context, base=address)

    def render_kernel_scan(self, renderer):
        modules = self.session.plugins.lsmod()

        for rule, address, _, _ in self.generate_hits(
                self.kernel_address_space):
            renderer.format("Rule: {0}\n", rule)

            # Find out who owns this hit.
            owner = modules.find_module(address)
            if owner:
                renderer.format("Owner: {0}\n", owner.name)
            else:
                renderer.format("Owner: (Unknown Kernel Memory)\n")

            context = self.kernel_address_space.read(address, 0x40)
            utils.WriteHexdump(renderer, context, base=address)

    def render_task_scan(self, renderer, task):
        task_as = task.get_process_address_space()
        end = self.session.GetParameter("highest_usermode_address")

        for rule, address, _, _ in self.generate_hits(task_as, end=end):
            renderer.format("Rule: {0}\n", rule)

            renderer.format("Owner: {0}\n", task.comm)

            context = task_as.read(address, 0x40)
            utils.WriteHexdump(renderer, context, base=address)


    def render(self, renderer):
        """Render output."""
        if self.scan_physical:
            return self.render_scan_physical(renderer)

        elif self.filtering_requested:
            for task in self.filter_processes():
                self.render_task_scan(renderer, task)

        # We are searching the kernel address space
        else:
            return self.render_kernel_scan(renderer)



# Rekall Memory Forensics
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""Various plugins that make entities testable and easier to profile."""

__author__ = "Adam Sindelar <adamsh@google.com>"


from rekall import plugin
from rekall import testlib

from rekall.plugins.common import entities


class DarwinOnlyMixin(object):
    @classmethod
    def is_active(cls, session):
        """Only active for Darwin right now."""
        return (session.profile.metadata("os") == "darwin" and
                plugin.Command.is_active(session))


class ListInterfaces(DarwinOnlyMixin, entities.EntityFind):
    name = "ifconfig"
    search = "has component NetworkInterface"

    columns = ["NetworkInterface/name",
               "NetworkInterface/endpoints->OSILayer2/address",
               "NetworkInterface/endpoints->OSILayer3/address"]

    sort = ["Endpoint/owner->NetworkInterface/name",
            "OSILayer3/protocol",
            "OSILayer3/address"]


class TestListInterfaces(testlib.SortedComparison):
    PARAMETERS = dict(commandline="ifconfig")


class ListZones(entities.EntityFind):
    name = "zones"
    search = "has component AllocationZone"
    columns = ["AllocationZone/name", "AllocationZone/count_active",
               "AllocationZone/count_free", "AllocationZone/element_size",
               "AllocationZone/tracks_pages", "AllocationZone/allows_foreign"]
    sort = ["AllocationZone/name"]


class TestListZones(testlib.SortedComparison):
    PARAMETERS = dict(commandline="zones")


class ListEvents(entities.EntityFind):
    name = "events"
    search = "has component Event"
    columns = ["Event/timestamp", "Event/category", "Event/actor",
               "Event/action", "Event/target"]
    sort = ["Event/timestamp", "Event/category", "Event/action",
            "Event/actor", "Event/target"]


class TestListEvents(testlib.SortedComparison):
    PARAMETERS = dict(commandline="events")

    def testCase(self):
        previous = sorted(set(self.baseline['output']))
        current = sorted(set(self.current['output']))

        self.assertListEqual(previous, current)


class LSOF(DarwinOnlyMixin, entities.EntityFind):
    __name = "lsof"
    description = "Open Files"
    width = 150
    search = ("Handle/resource matches "
              "(has component File or has component Connection or "
              "has component Struct) "
              "and Handle/process->Process/user matches "
              "(has component User)")
    columns = ["Handle/process", "Handle/process->Process/user", "Handle/fd",
               "Handle/resource->Named/kind",
               "Handle/resource"]
    sort = ["Handle/process->Process/pid", "Handle/fd"]


class DarwinListFiles(DarwinOnlyMixin, entities.EntityFind):
    __name = "list_files"
    description = "All Files"
    width = 110
    search = "has component File"
    columns = [dict(width=10, attribute="File/type"),
               dict(name="Sources",
                    fn=lambda e: ",".join(sorted(e["Entity/collectors"])),
                    width=30),
               dict(name="Created", width=15,
                    attribute="Timestamps/created_at"),
               dict(name="Modified", width=15,
                    attribute="Timestamps/modified_at"),
               dict(attribute="File/path", width=40)]
    sort = ["Timestamps/created_at", "Timestamps/modified_at"]


class TestDarwinListFiles(testlib.SortedComparison):
    PARAMETERS = dict(commandline="list_files")


class IPNetstat(entities.EntityFind):
    __name = "ipnetstat"
    description = "IP Connections"
    width = 150
    search = "Connection/protocol_family in ('INET', 'INET6')"
    columns = [dict(name="protocol",
                    fn=lambda e: "/".join([
                        e["Connection/source->OSILayer3/protocol"],
                        e["Connection/source->OSILayer4/protocol"]])),
               dict(attribute="Connection/source->OSILayer3/address",
                    name="src address"),
               dict(attribute="Connection/source->OSILayer4/port",
                    name="src port"),
               dict(attribute="Connection/destination->OSILayer3/address",
                    name="dst address"),
               dict(attribute="Connection/destination->OSILayer4/port",
                    name="dst port"),
               "Connection/source->OSILayer4/state",
               "Connection/handles->Handle/process"]
    sort = ["Connection/source->OSILayer3/protocol",
            "Connection/source->OSILayer4/protocol",
            "Connection/destination->OSILayer3/address",
            "Connection/destination->OSILayer4/port",
            "Connection/source->OSILayer3/address",
            "Connection/source->OSILayer4/port"]


class TestIPNetstat(testlib.SortedComparison):
    PARAMETERS = dict(commandline="ipnetstat")


class SocketNetstat(DarwinOnlyMixin, entities.EntityFind):
    __name = "unix_sockets"
    description = "UNIX sockets"
    search = "Connection/protocol_family is 'UNIX'"
    columns = ["Socket/type", "Socket/address", "Socket/connected",
               "Socket/file", "Connection/handles->Handle/process"]
    sort = ["Socket/address"]


class EntityNetstat(DarwinOnlyMixin, entities.FindBatch):
    __name = "netstat"
    batch = ["ipnetstat", "unix_sockets"]



# Rekall Memory Forensics
# Copyright (C) 2007,2008 Volatile Systems
# Copyright (C) 2010,2011,2012 Michael Hale Ligh <michael.ligh@mnin.org>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""The following is a description of windows stations from MSDN:

http://msdn.microsoft.com/en-us/library/windows/desktop/ms687096(v=vs.85).aspx

A window station contains a clipboard, an atom table, and one or more desktop
objects. Each window station object is a securable object. When a window station
is created, it is associated with the calling process and assigned to the
current session.

The interactive window station is the only window station that can display a
user interface or receive user input. It is assigned to the logon session of the
interactive user, and contains the keyboard, mouse, and display device. It is
always named "WinSta0". All other window stations are noninteractive, which
means they cannot display a user interface or receive user input.


Ref:
http://volatility-labs.blogspot.de/2012/09/movp-13-desktops-heaps-and-ransomware.html

NOTE: Windows 8 does not have a global atom table any more.
http://mista.nu/research/smashing_the_atom.pdf

"""
from rekall import plugin
from rekall.plugins.windows import common
from rekall.plugins.windows.gui import win32k_core


class WindowsStations(win32k_core.Win32kPluginMixin,
                      common.WindowsCommandPlugin):
    """Displays all the windows stations by following lists."""

    __name = "windows_stations"

    def stations_in_session(self, session):
        # Get the start of the Window station list from
        # win32k.sys. These are all the Windows stations that exist in
        # this Windows session.
        station_list = self.win32k_profile.get_constant_object(
            "grpWinStaList",
            target="Pointer",
            target_args=dict(
                target="tagWINDOWSTATION"
                ),
            vm=session.obj_vm,
            )

        for station in station_list.walk_list("rpwinstaNext"):
            yield station

    def stations(self):
        """A generator of tagWINDOWSTATION objects."""
        # Each windows session has a unique set of windows stations.
        for session in self.session.plugins.sessions().session_spaces():
            for station in self.stations_in_session(session):
                yield station

    def render(self, renderer):
        for window_station in self.stations():
            renderer.section()
            renderer.format(
                "WindowStation: {0:addr}, Name: {1}, Next: {2:addr}\n",
                window_station,
                window_station.Name,
                window_station.rpwinstaNext)

            renderer.format(
                "SessionId: {0}, AtomTable: {1:addr}, "
                "Interactive: {2}\n",
                window_station.dwSessionId,
                window_station.pGlobalAtomTable,
                window_station.Interactive)

            renderer.format(
                "Desktops: {0}\n",
                [desk.Name for desk in window_station.desktops()])

            ethread = window_station.ptiDrawingClipboard.pEThread

            renderer.format(
                "ptiDrawingClipboard: pid {0} tid {1}\n",
                ethread.Cid.UniqueProcess, ethread.Cid.UniqueThread)

            last_registered_viewer = window_station.LastRegisteredViewer
            renderer.format(
                "spwndClipOpen: {0:addr}, spwndClipViewer: {1:addr} {2} {3}\n",
                window_station.spwndClipOpen,
                window_station.spwndClipViewer,
                last_registered_viewer.UniqueProcessId,
                last_registered_viewer.ImageFileName)

            renderer.format("cNumClipFormats: {0}, iClipSerialNumber: {1}\n",
                            window_station.cNumClipFormats,
                            window_station.iClipSerialNumber)

            renderer.format(
                "pClipBase: {0:addr}, Formats: {1}\n",
                window_station.pClipBase,
                [clip.fmt for clip in window_station.pClipBase.dereference()])



class WinDesktops(plugin.VerbosityMixIn, WindowsStations):
    """Print information on each desktop."""

    __name = "desktops"

    def render(self, renderer):
        for window_station in self.stations():
            for desktop in window_station.desktops():
                renderer.section()

                renderer.format(
                    "Desktop: {0:addr}, Name: {1}\\{2}, Next: {3:addr}\n",
                    desktop,
                    window_station.Name,
                    desktop.Name,
                    desktop.rpdeskNext.v(),
                    )

                renderer.format(
                    "SessionId: {0}, DesktopInfo: {1:addr}, fsHooks: {2}\n",
                    desktop.dwSessionId,
                    desktop.pDeskInfo.v(),
                    desktop.DeskInfo.fsHooks,
                    )

                renderer.format(
                    "spwnd: {0:addr}, Windows: {1}\n",
                    desktop.DeskInfo.spwnd,
                    len(list(desktop.windows(desktop.DeskInfo.spwnd)))
                    )
                renderer.format(
                    "Heap: {0:addr}, Size: {1:addr}, Base: {2:addr}, "
                    "Limit: {3:addr}\n",
                    desktop.pheapDesktop.v(),
                    (desktop.DeskInfo.pvDesktopLimit.v() -
                     desktop.DeskInfo.pvDesktopBase.v()),
                    desktop.DeskInfo.pvDesktopBase,
                    desktop.DeskInfo.pvDesktopLimit,
                    )

                # Print heap allocations.
                if self.verbosity > 1:
                    for entry in desktop.pheapDesktop.Entries:
                        renderer.format(
                            "   Alloc: {0:addr}, Size: {1:addr} "
                            "Previous: {2:addr}\n",
                            entry,
                            entry.Size,
                            entry.PreviousSize
                            )

                for thrd in desktop.threads():
                    renderer.format(
                        " {0} ({1} {2} parent {3})\n",
                        thrd.pEThread.Cid.UniqueThread,
                        thrd.ppi.Process.ImageFileName,
                        thrd.ppi.Process.UniqueProcessId,
                        thrd.ppi.Process.InheritedFromUniqueProcessId,
                        )



# Rekall Memory Forensics
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
The Rekall Entity Layer.
"""
__author__ = "Adam Sindelar <adamsh@google.com>"

import logging
import traceback

from rekall import obj

from rekall.entities import component as entity_component
from rekall.entities import identity
from rekall.entities import types

from efilter import query

from efilter.protocols import superposition


class CurriedComponent(object):
    component = None
    entity = None

    def __init__(self, entity, component):
        self.component = component
        self.entity = entity

    def get_raw(self, key):
        return self.component[key]

    def get(self, key, complete=False):
        return self.entity.get(
            key="%s/%s" % (self.component.component_name, key),
            complete=complete)

    def __getitem__(self, key):
        return self.get(key)

    def __repr__(self):
        return "CurriedComponent(component=%r, entity=%r)" % (
            self.component, self.entity)


class Entity(object):
    """Entity is an abstraction of things like Process, User or Connection.

    Each entity consists of a an identity object, like PID, username or a memory
    address, and a list of components, which are data about the entity.

    Immutability:
    =============

    Entities and components are copy-on-write. Components enforce immutability
    while Entity and Identity are less strict for practical reasons.
    Nevertheless, changing an instance of entity or identity after they've been
    added to an entity manager will result in undefined behavior.

    Relational Model:
    =================

    Entites support associations with other entities. For example, a resource
    has a handle, a handle is owned by a process, and so on. These associations
    are implemented using Identity and lookups through an entity manager. See
    documentation of 'find_referenced_entities' and 'find_referencing_entities'
    for a discussion of associations.

    Merging:
    ========

    If more than one entity is added to an entity manager for the same thing,
    they will be merged into a single entity with all the properties, as long as
    the identity is decidable. For example, if we have a user with UID 15 and
    another user with username "Alice" they will remain separate entities until
    an entity is added with an identity that has the UID 15 and username
    "Alice", at which point the identity becomes decidable and all three
    entities will be merged into one.

    If, during merging, a conflict arises then both versions are kept
    encapsulated in an instance of superposition.Superposition. As long as you
    use 'get_' family of accessors on Entity this will be handled for you (as
    those functions always return generators).

    Superpositions:
    ===============

    Superpositions are technically merge conflicts, but they are not errors.
    They occur in cases where there are legitimately multiple valid values of a
    single attribute. Some examples:

    Resource/handle can have multiple values, because a single file/socket can
    be opened by multiple handles owned by multiple processes.

    Struct/type can be a superposition in case of unions, or things that
    are stored as a void pointer and cast depending on contextual state.

    User/real_name can often be a superposition because of variable formatting
    rules applied by the OS.

    Public state members:
    =====================

    components: An instance of ComponentContainer, which is a fast container
        object that actually stores the components of the entity.

    manager: The manager this entity belongs to.

    Special Attributes:
    ===================

    entity["Entity/identity"] (shorthand entity.identity, entity.indices)
    entity["Entity/collectors"] (shorthand entity.collectors)

    ### References:

    http://en.wikipedia.org/wiki/Composition_over_inheritance
    http://en.wikipedia.org/wiki/Entity_component_system
    """

    typedesc = None

    def __init__(self, components, entity_manager=None):
        self.components = components
        self.manager = entity_manager
        self.typedesc = EntityDescriptor()

    @property
    def identity(self):
        return self.components.Entity.identity

    @property
    def collectors(self):
        return self.components.Entity.collectors

    @property
    def indices(self):
        """Returns all the keys that the entity will be accessible at."""
        return self.identity.indices

    def __hash__(self):
        return hash(self.identity)

    def __eq__(self, other):
        if not isinstance(other, Entity):
            return False

        return self.identity == other.identity

    def __ne__(self, other):
        return not self.__eq__(other)

    def issuperset(self, other):
        """Is this entity a strict superset of other?"""
        # Compare both component containers, but skip the Entity component.
        other_components = other.components[1:]
        for cidx, component in enumerate(self.components[1:]):
            other_comp = other_components[cidx]
            if other_comp is None:
                continue

            if component is None or not component.issuperset(other_comp):
                return False

        return True

    @property
    def name(self):
        """Name of the entity, if set (Named/name). If not, takes a guess."""
        name = self.get_raw("Named/name")

        if name:
            try:
                return unicode(name)
            except UnicodeDecodeError:
                logging.warning(
                    "Could not decode Named/name '%s' as unicode.",
                    name.encode("string-escape"))
                name = None

        if not name:
            key = unicode(self.identity.first_index[1])
            val = self.identity.first_index[2]

            if isinstance(val, obj.Struct):
                # Rekall uses the opposite meaning of repr and str from
                # the entity layer. This is a temporary workaround until
                # everything just uses renderers all the time.
                val = repr(val)
            else:
                val = unicode(val)

            return "%s: %s" % (key, val)

        return unicode(name)

    @property
    def kind(self):
        kind = self.get_raw("Named/kind")
        if kind == None:
            kind = "Entity"

        return unicode(kind)

    def __repr__(self):
        parts = []
        for component in self.components:
            if not component:
                continue

            parts.append(repr(component))

        return "Entity(\n%s)" % ",\n\n".join(parts)

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return u"%s: %s" % (self.kind, self.name)

    # pylint: disable=protected-access
    def asdict(self):
        """Returns a dict of all attributes and their values."""
        result = {}
        for component_name in entity_component.Component.classes.keys():
            component = getattr(self.components, component_name)

            if component is None:
                continue

            for idx, field in enumerate(component.component_fields):
                val = component[idx]
                if val:
                    key = "%s/%s" % (component_name, field.name)
                    result[key] = val

        return result

    def export(self):
        """Returns a dict of all attribute values with extra descriptions."""
        result = {}
        for component_name in entity_component.Component.classes.keys():
            component = getattr(self.components, component_name)

            if component is None:
                continue

            for idx, field in enumerate(component.component_fields):
                val = component[idx]

                if val is None:
                    continue

                if field.typedesc.type_name == "Identity":
                    entity = self.manager.find_by_identity(val)
                    if entity:
                        val.name = entity[0].name
                        val.kind = entity[0].kind

                key = "%s/%s" % (component_name, field.name)
                result[key] = val

        return result

    @classmethod
    def fromdict(cls, attributes, manager=None):
        """Rebuilds the entity from a dict of attributes.

        Values will be coerced into their proper types automatically.
        """
        component_dicts = {}

        for attribute, value in attributes.iteritems():
            component_name, field_name = attribute.split("/", 1)
            component_dicts.setdefault(component_name, {})
            component_dicts[component_name][field_name] = value

        components = {}
        for component_name, kwargs in component_dicts.iteritems():
            component_cls = cls.reflect_component(component_name)
            components[component_name] = component_cls(**kwargs)

        return cls(entity_component.CONTAINER_PROTOTYPE._replace(**components),
                   manager)

    def get_referencing_entities(self, key, complete=True):
        """Finds entities that reference this entity by its identity.

        If other entities have an attribute that stores the identity of this
        entity then this function will find those entities.

        For example, calling this on a process, one can find all the handles
        owned by the process by calling
        process.get_referencing_entities("Handle/process").

        Arguments:
            key: The property path to the attribute on the other entities.
                As usual, form is Component/attribute.
        """
        # Automatically ask the entity manager to add indexing for
        # identity-based attributes. This is a good heuristic for optimal
        # performance.
        self.manager.add_attribute_lookup(key)

        return self.manager.find(
            query.Query("%s is {}" % key, params=[self.identity]),
            complete=complete, keep_cache=True)

    def get_raw(self, key):
        """Get raw value of the key, no funny bussiness.

        Does not attempt to resolve superposition or identities, just returns
        the raw value of the key.

        Arguments:
            key: Property path in form of Component/attribute.
        """
        try:
            component_name, attribute = key.split("/", 1)
        except ValueError:
            # Maybe the key is just a component name?
            component = getattr(self.components, key, None)
            if component:
                return component

            raise ValueError("%s is not a valid key." % key)

        component = getattr(self.components, component_name, None)
        if component is None:
            return None

        try:
            return getattr(component, attribute)
        except AttributeError:
            return component.reflect_attribute(attribute)

    def get(self, key, complete=False):
        """Returns value of the key, or a superposition thereof.

        Out of the get_ functions, this is almost always the one you want.

        Getting a basic value:
        ======================

        Use key in form of Component.attribute. For example, "Process/pid" or
        "User/username". Same as calling entity[key]:

        entity["Process/pid"]  # PID of the process.

        What if the value is an entity:
        ===============================

        This method automatically recognizes attributes that reference other
        entities, looks them up and returns them. For example:

        entity["Process/parent"]  # Returns the parent process entity.
        entity["Process/parent"]["Process/pid"]  # PID of the parent.

        What if I want all the child processes (Inverse Lookup):
        ========================================================

        You can call entity.get_referencing_entities if you want to be explicit.

        Alternatively, prepend the key with a '&' for inverse lookup of a N:1
        assocation.

        For example:

        entity["&Process/parent"]  # Returns processes of which this process is
                                   # (Child processes).

        entity["&Handle/process"]  # Returns all handles this process has open.

        In most cases, this is unnecessary and aliases can be used instead.

        For example:

        entity["&Handle/process"]  # Is already aliased in definitions as:
        entity["Process/handles"]

        entity["&Process/parent"]  # Is already aliased in definitions as:
        entity["Process/children"]

        When does this return more than one value:
        ==========================================

        1) When doing an inverse lookup.
        2) When the value we find is a superposition.

        In both cases, a superposition is returned. Remember that superpositions
        proxy the [] operator, returning more superpositions. For example:

        # To return the pids of all child processes:
        entity["&Process/parent"]["Process/pid"]

        You can request more than one key in a single call:
        ===================================================

        This is identical to the behavior of [] on python dictionaries:

        entity["Process/pid", "Process/command"]  # is the same as calling:
        (entity["Process/pid"], entity["Process/command"])

        You can also request a multi-level path into the object:
        ========================================================

        The path separator is '->' and is used as follows:

        entity["Process/handles"]["Handle/resource"]  # Is the same as:
        entity["Process/handles->Handle/resource]

        This is merely syntax sugar to make specifying rendering output easier.
        """
        # If we get called with [x, y] python will pass us the keys as a tuple
        # of (x, y). The following behaves identically to dict.
        if isinstance(key, tuple):
            return [self.get(_key, complete) for _key in key]

        # If we get called with a Component/attribute->Component/attribute
        # we treat -> as path separator and the whole key as path into the
        # object.
        if "->" in key:
            # This works recursively.
            key, rest = key.split("->", 1)
            subresult = self.get(key=key, complete=complete)
            if not subresult:
                return None

            return superposition.state_apply(
                subresult,
                lambda x: x.get(key=rest, complete=complete))

        # The & sigil denotes reverse lookup.
        if key.startswith("&"):
            entities = self.get_referencing_entities(key[1:],
                                                     complete=complete)
            return superposition.meld(*list(entities))

        # The raw result could be None, a superposition or just a scalar.
        value = self.get_raw(key)
        if value is None:
            return obj.NoneObject(
                "Entity '%s' has no results for key '%s'." % (self, key))

        # Redirection.
        if isinstance(value, entity_component.Alias):
            return self.get(value.alias, complete=complete)

        if isinstance(value, entity_component.Component):
            return CurriedComponent(self, value)

        typedesc = self.reflect_type(key)
        if typedesc.type_name == "Entity":
            entities = self.manager.find_by_identity(value, complete=complete)
            return superposition.meld(*list(entities))

        return value

    def get_variants(self, key, complete=False):
        value = self.get(key, complete=complete)
        return superposition.getstates(value)

    def __getitem__(self, key):
        return self.get(key)

    @classmethod
    def reflect_attribute(cls, path):
        """Return an instance of Attribute describing the attribute."""
        # For longer keypaths, the relevant attribute (for types, etc) is
        # the last one.
        attribute_name = path.split("->")[-1]

        try:
            component, key = attribute_name.split("/", 1)
        except ValueError:
            # Doesn't include a slash - it's a component maybe?
            return cls.reflect_component(path)

        component_cls = cls.reflect_component(component)
        if not component_cls:
            return

        attribute = component_cls.reflect_attribute(key)

        if "->" in path:
            # Need to modify the attribute so that attribute.path matches the
            # one the callers was looking for.
            return entity_component.AttributePath(attribute, path)

        return attribute

    @classmethod
    def reflect_type(cls, path):
        typedesc = cls.reflect_attribute(path).typedesc
        if typedesc.type_name == "Identity":
            # Return a typedesc that says "Entity" on it when asked. Even
            # though the attribute contains an Identity, and that's what will
            # be set by any writers, Entity is what we actually return from
            # self.get.
            return EntityDescriptor()

        return typedesc

    @classmethod
    def reflect_component(cls, component):
        return entity_component.Component.classes.get(component, None)

    def _merge_containers(self, other):
        """Merge component containers from self and other into new container."""
        new_components = []
        x = self.components
        y = other.components

        # Skipping component idx 0 (Entity)
        try:
            for idx in xrange(1, len(x)):
                cx = x[idx]
                cy = y[idx]
                if not cx:
                    if cy:
                        new_components.append(cy)
                    else:
                        new_components.append(None)
                else:
                    if cy:
                        new_components.append(cx.union(cy))
                    else:
                        new_components.append(cx)
        except TypeError as e:
            # Merging may encounter type enforcement errors. Such errors mean
            # either that a collector is using faulty assumptions (canonical
            # example is assuming handles on Windows are unique to a process,
            # and using the process identity in the Handle component) or that
            # the image is inconsistent.
            # The best we can do is reject the faulty data and hope for the
            # best.
            logging.warn("Invalid entity (type error) encountered. "
                         "Run with debug logging for details.")
            logging.debug("REJECTED entity because of type error. "
                          "Dump of entity and traceback follow: \n"
                          "%r\n%r\n%s",
                          other, e, traceback.format_exc())
            return self.components

        # Entity component is merged using slightly simpler rules.
        component_cls = entity_component.Component.classes["Entity"]
        try:
            new_entity_component = component_cls(
                identity=x.Entity.identity | y.Entity.identity,
                collectors=x.Entity.collectors | y.Entity.collectors)
        except identity.IdentityError as e:
            # Consistency errors can mean the data is corrupt. There are no
            # good solutions. Lets log the reject and get on with our lives.
            logging.warn("Invalid entity (logic error) encountered. "
                         "Run with debug logging for details.")
            logging.debug("REJECTED entity because of logic error. "
                          "Dump of entity and traceback follow: \n"
                          "%r\n%r\n%s",
                          other, e, traceback.format_exc())
            return self.components

        return type(x)(new_entity_component, *new_components)

    def update(self, other):
        """Changes this entity to include information from other.

        This is not a part of the API - only EntityManager should use this.
        """
        self.components = self._merge_containers(other)

    def union(self, other):
        """Returns a new entity that is a union of x and y.

        Original entities remain immutable. If x and y have some of the same
        components and those components report conflicting values in some fields
        then those fields will be replaced with a superposition (see
        superposition.Superposition) of all reported values.
        """
        if not self == other:  # self != other can return None.
            raise AttributeError("Can't do union unless both are equal.")

        return Entity(
            components=self._merge_containers(other),
            entity_manager=self.manager)

    def __ior__(self, other):
        return self.update(other)

    def __or__(self, other):
        return self.union(other)


class IdentityDescriptor(types.TypeDescriptor):
    """Entity type, actually contains the Identity key."""

    # Actually, it's Identity, but the user doesn't care about the difference.
    type_name = "Identity"

    def coerce(self, value):
        if value is None:
            return None

        if isinstance(value, Entity):
            return value.identity

        if isinstance(value, identity.Identity):
            return value

        value_repr = None
        try:
            value_repr = repr(value)
        except Exception as e:
            raise TypeError(
                ("Object passed to coerce is not an identity. Additionally, "
                 "calling repr(object) raised %s.") % e)

        if superposition.insuperposition(value):
            if superposition.state_type(value) == identity.Identity:
                return value

            raise TypeError(
                "Object being coerced as identity is a superposition %s of "
                "variant type %r." % (value_repr, value.type_name))

        raise TypeError("%s is not an identity." % value_repr)

    def __repr__(self):
        return "Entity (Identity) type"


class EntityDescriptor(types.TypeDescriptor):
    """Actually contains an Entity. Used by superpositions."""

    type_name = "Entity"
    type_cls = Entity

    def coerce(self, value):
        if value == None:
            return value

        if isinstance(value, Entity):
            return value

        value_repr = None
        try:
            value_repr = repr(value)
        except Exception as e:
            raise TypeError(
                ("Object passed to coerce is not an entity. Additionally, "
                 "calling repr(object) raised %s.") % e)

        raise TypeError("%s is not an entity." % value_repr)

    def __repr__(self):
        return "Entity type"



# Import and register the core plugins
# pylint: disable=unused-import

from rekall.plugins.renderers import base_objects
from rekall.plugins.renderers import darwin
from rekall.plugins.renderers import data_export
from rekall.plugins.renderers import entities
from rekall.plugins.renderers import json_storage
from rekall.plugins.renderers import linux
from rekall.plugins.renderers import virtualization
from rekall.plugins.renderers import visual_aides
from rekall.plugins.renderers import windows

try:
    from rekall.plugins.renderers import xls
except ImportError:
    pass



# Rekall Memory Forensics
# Copyright (C) 2007,2008 Volatile Systems
# Copyright (c) 2008 Brendan Dolan-Gavitt <bdolangavitt@wesleyan.edu>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Additional Authors:
# Mike Auty <mike.auty@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

# pylint: disable=protected-access

import os
import re

from rekall.plugins.windows import common
from rekall.plugins import core
from rekall import utils


class PEDump(common.WindowsCommandPlugin):
    """Dump a PE binary from memory."""

    __name = "pedump"

    @classmethod
    def args(cls, parser):
        """Declare the command line args we need."""
        super(PEDump, cls).args(parser)
        parser.add_argument(
            "--image_base", default=0, type="IntParser",
            help="The address of the image base (dos header).")

        parser.add_argument("--out_file", default=None,
                            help="The file name to write.")

        parser.add_argument("-a", "--address_space", default=None,
                            help="The address space to use.")

        parser.add_argument(
            "--out_fd", dest="SUPPRESS",
            help="A file like object to write the output.")

    def __init__(self, image_base=0, out_fd=None, address_space=None,
                 out_file=None, **kwargs):
        super(PEDump, self).__init__(**kwargs)
        self.address_space = address_space

        # Allow the image base to be given as a name (e.g. "nt").
        resolver = self.session.address_resolver
        if resolver:
            image_base = resolver.get_address_by_name(image_base)

        self.image_base = image_base
        self.out_fd = out_fd
        self.out_file = out_file

        # Get the pe profile.
        self.pe_profile = self.session.LoadProfile("pe")

    def WritePEFile(self, fd=None, address_space=None, image_base=None):
        """Dumps the PE file found into the filelike object.

        Note that this function can be used for any PE file (e.g. executable,
        dll, driver etc). Only a base address need be specified. This makes this
        plugin useful as a routine in other plugins.

        Args:
          fd: A writable filelike object which must support seeking.
          address_space: The address_space to read from.
          image_base: The offset of the dos file header.
        """
        dos_header = self.pe_profile._IMAGE_DOS_HEADER(
            offset=image_base, vm=address_space)

        image_base = dos_header.obj_offset
        nt_header = dos_header.NTHeader

        # First copy the PE file header, then copy the sections.
        data = dos_header.obj_vm.read(
            image_base, min(1e6, nt_header.OptionalHeader.SizeOfHeaders))

        if not data:
            return

        fd.seek(0)
        fd.write(data)

        for section in nt_header.Sections:
            # Force some sensible maximum values here.
            size_of_section = min(10e6, section.SizeOfRawData)
            physical_offset = min(100e6, int(section.PointerToRawData))

            data = section.obj_vm.read(
                section.VirtualAddress + image_base, size_of_section)

            fd.seek(physical_offset, 0)
            fd.write(data)

    def render(self, renderer):
        if self.out_file:
            self.out_fd = renderer.open(filename=self.out_file, mode="wb")

        if not self.out_fd:
            self.session.logging.error(
                "No output filename or file handle specified.")
            return

        # Default address space is the kernel if not specified.
        if self.address_space is None:
            self.address_space = self.session.GetParameter(
                "default_address_space")

        with self.out_fd:
            image_base = self.session.address_resolver.get_address_by_name(
                self.image_base)
            renderer.format("Dumping PE File at image_base {0:#x} to {1}\n",
                            image_base, self.out_file)

            self.WritePEFile(self.out_fd, self.address_space, self.image_base)

            renderer.format("Done!\n")


class ProcExeDump(core.DirectoryDumperMixin, common.WinProcessFilter):
    """Dump a process to an executable file sample"""

    __name = "procdump"

    dump_dir_optional = True

    @classmethod
    def args(cls, parser):
        super(ProcExeDump, cls).args(parser)

        parser.add_argument(
            "--out_fd", dest="SUPPRESS",
            help="A file like object to write the output.")

    def __init__(self, out_fd=None, **kwargs):
        """Dump a process from memory into an executable.

        In windows PE files are mapped into memory in sections. Each section is
        mapped into a region within the process virtual memory from a region in
        the executable file:

    File on Disk                 Memory Image
0-> ------------    image base-> ------------
     Header                      Header
    ------------                 ------------
     Section 1
    ------------                 ------------
     Section 2                    Section 1
    ------------                 ------------

                                 ------------
                                  Section 2
                                 ------------

        This plugin simply copies the sections from memory back into the file on
        disk. Its likely that some of the pages in memory are not actually
        memory resident, so we might get invalid page reads. In this case the
        region on disk is null padded. If that happens it will not be possible
        to run the executable, but the executable can still be disassembled and
        analysed statically.

        References:
        http://code.google.com/p/corkami/downloads/detail?name=pe-20110117.pdf

        NOTE: Malware can mess with the headers after loading. The remap option
        allows to remap the sections on the disk file so they do not collide.

        Args:
          remap: If set, allows to remap the sections on disk so they do not
            overlap.

          out_fd: Alternatively, a filelike object can be provided directly.
        """
        # If a fd was not provided, the dump_dir must be specified.
        if out_fd is None:
            self.dump_dir_optional = False

        super(ProcExeDump, self).__init__(**kwargs)
        self.fd = out_fd
        self.pedump = PEDump(session=self.session)

    def render(self, renderer):
        """Renders the tasks to disk images, outputting progress as they go"""
        for task in self.filter_processes():
            pid = task.UniqueProcessId

            task_address_space = task.get_process_address_space()
            if not task_address_space:
                renderer.format("Can not get task address space - skipping.")
                continue

            if self.fd:
                self.pedump.WritePEFile(
                    self.fd, task_address_space, task.Peb.ImageBaseAddress)
                renderer.section()

                renderer.format("Dumping {0}, pid: {1:6} into user provided "
                                "fd.\n", task.ImageFileName, pid)

            # Create a new file.
            else:
                sanitized_image_name = re.sub(
                    "[^a-zA-Z0-9-_]", "_", utils.SmartStr(task.ImageFileName))

                filename = u"executable.%s_%s.exe" % (
                    sanitized_image_name, pid)

                renderer.section()

                renderer.format("Dumping {0}, pid: {1:6} output: {2}\n",
                                task.ImageFileName, pid, filename)

                with renderer.open(directory=self.dump_dir,
                                   filename=filename,
                                   mode="wb") as fd:
                    # The Process Environment Block contains the dos header:
                    self.pedump.WritePEFile(
                        fd, task_address_space, task.Peb.ImageBaseAddress)


class DLLDump(ProcExeDump):
    """Dump DLLs from a process address space"""

    __name = "dlldump"

    @classmethod
    def args(cls, parser):
        """Declare the command line args we need."""
        super(DLLDump, cls).args(parser)
        parser.add_argument(
            "--regex", default=".+",
            help="A Regular expression for selecting the dlls to dump.")

    def __init__(self, regex=".+", **kwargs):
        """Dumps dlls from processes into files.

        Args:
          regex: A regular expression that is applied to the modules name.
        """
        super(DLLDump, self).__init__(**kwargs)
        self.regex = re.compile(regex)

    def render(self, renderer):
        renderer.table_header([("_EPROCESS", "eprocess", "[addrpad]"),
                               ("Name", "name", "16"),
                               ("Base", "base", "[addrpad]"),
                               ("Module", "module", "20s"),
                               ("Dump File", "filename", "")])

        for task in self.filter_processes():
            task_as = task.get_process_address_space()

            # Skip kernel and invalid processes.
            for module in task.get_load_modules():
                process_offset = task_as.vtop(task.obj_offset)
                if process_offset:

                    # Skip the modules which do not match the regex.
                    if not self.regex.search(
                            utils.SmartUnicode(module.BaseDllName)):
                        continue

                    base_name = os.path.basename(
                        utils.SmartUnicode(module.BaseDllName))

                    dump_file = "module.{0}.{1:x}.{2:x}.{3}".format(
                        task.UniqueProcessId, process_offset, module.DllBase,
                        base_name)

                    renderer.table_row(
                        task, task.name, module.DllBase, module.BaseDllName,
                        dump_file)

                    # Use the procdump module to dump out the binary:
                    with renderer.open(filename=dump_file,
                                       directory=self.dump_dir,
                                       mode="wb") as fd:
                        self.pedump.WritePEFile(fd, task_as, module.DllBase)

                else:
                    renderer.format(
                        "Cannot dump {0}@{1} at {2:8x}\n",
                        task.ImageFileName, module.BaseDllName, module.DllBase)


class ModDump(DLLDump):
    """Dump kernel drivers from kernel space."""

    __name = "moddump"

    address_spaces = None

    def find_space(self, image_base):
        """Search through all process address spaces for a PE file."""
        if self.address_spaces is None:
            self.address_spaces = [self.kernel_address_space]
            for task in self.filter_processes():
                self.address_spaces.append(task.get_process_address_space())

        for address_space in self.address_spaces:
            if address_space.is_valid_address(image_base):
                return address_space

    def render(self, renderer):
        modules_plugin = self.session.plugins.modules(session=self.session)

        for module in modules_plugin.lsmod():
            if self.regex.search(utils.SmartUnicode(module.BaseDllName)):
                address_space = self.find_space(module.DllBase)
                if address_space:
                    dump_file = "driver.{0:x}.sys".format(module.DllBase)
                    renderer.format("Dumping {0}, Base: {1:8x} output: {2}\n",
                                    module.BaseDllName, module.DllBase,
                                    dump_file)

                    with renderer.open(filename=dump_file,
                                       directory=self.dump_dir,
                                       mode="wb") as fd:
                        self.pedump.WritePEFile(
                            fd, address_space, module.DllBase)
